from unittest.mock import MagicMock

import pytest
from fastapi.requests import Request


class MockRequest:
    """Request cls required to test pagination."""

    class State:
        def __init__(self):
            self.paginate_by = None

    def __init__(self):
        self.state = self.State()
        self.query_params = None
        self.url = None


@pytest.fixture()
def request_fixture():
    mock = MagicMock(spec=Request, return_value=MockRequest())
    mock.url = 'https://example.com'
    return mock



import pytest
from fastapi.exceptions import HTTPException

from fastapi_views.ext.pymongo.pagination.core import PaginationCursor
from fastapi_views.ext.pymongo.sync_api import SyncAPI

NUMBER_OF_OBJECTS = 50  # number of objects created by sync db


@pytest.fixture
def pymongo_api(request):
    collection, pk_field, paginate_by, strategy, kwargs, statement = request.param
    return SyncAPI(
        model=collection,
        pk_field=pk_field,
        paginate_by=paginate_by,
        pagination_strategy=strategy,
        pagination_kwargs=kwargs,
        statement=statement,
    )


class TestSyncAPI:
    """Test pymongo api cls."""

    @pytest.mark.parametrize('test_input', [None, ''])
    def test_init_model_raise_value_error(self, test_input):
        """Test collection value error."""
        with pytest.raises(ValueError):
            assert SyncAPI(model=test_input)

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_session_get_all_raises(self, pymongo_api: SyncAPI):
        """Test session attr in get_all method."""
        with pytest.raises(TypeError):
            pymongo_api.get_all(db=None, session=False)
        with pytest.raises(TypeError):
            pymongo_api.get_all(db=None, session=True)
        with pytest.raises(TypeError):
            pymongo_api.get_all(db=None, session='foo_str')

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_session_get_detail_raises(self, pymongo_api: SyncAPI):
        """Test session attr in get_detail method."""
        with pytest.raises(TypeError):
            pymongo_api.get_detail(db=None, session=False)
        with pytest.raises(TypeError):
            pymongo_api.get_detail(db=None, session=True)
        with pytest.raises(TypeError):
            pymongo_api.get_detail(db=None, session='foo_str')

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_session_update_one_raises(self, pymongo_api: SyncAPI):
        """Test session attr in update_one method."""
        with pytest.raises(TypeError):
            pymongo_api.update_one(db=None, session=False)
        with pytest.raises(TypeError):
            pymongo_api.update_one(db=None, session=True)
        with pytest.raises(TypeError):
            pymongo_api.update_one(db=None, session='foo_str', update_doc={})

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_session_create_raises(self, pymongo_api: SyncAPI):
        """Test session attr in create method."""
        with pytest.raises(TypeError):
            pymongo_api.create(db=None, session=False, document={})
        with pytest.raises(TypeError):
            pymongo_api.create(db=None, session=True, document={})
        with pytest.raises(TypeError):
            pymongo_api.create(db=None, session='foo_str', document={})

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_session_delete_raises(self, pymongo_api: SyncAPI):
        """Test session attr in delete method."""
        with pytest.raises(TypeError):
            pymongo_api.delete(db=None, session=False, field_value=None)
        with pytest.raises(TypeError):
            pymongo_api.delete(db=None, session=True, field_value=None)
        with pytest.raises(TypeError):
            pymongo_api.delete(db=None, session='foo_str', field_value=None)

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_get_all_statement_raises(self, pymongo_api: SyncAPI):
        """Test statement attr in get_all method."""
        with pytest.raises(TypeError):
            class Cls:
                pass

            pymongo_api.get_all(db=None, session=Cls)
        with pytest.raises(TypeError):
            pymongo_api.get_all(db=None, session=False)
        with pytest.raises(TypeError):
            pymongo_api.get_all(db=None, session=True)
        with pytest.raises(TypeError):
            pymongo_api.get_all(db=None, session='foo_str')
        with pytest.raises(TypeError):
            pymongo_api.get_all(db=None, session=2)

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, None, None, None, {})],
        indirect=['pymongo_api'],
    )
    def test_session_get_methods(self, pymongo_api: SyncAPI, pymongo_database, client_session):
        """Test session correct type. In some cases session is not implemented in mongomock yet."""
        assert pymongo_api.get_all(db=pymongo_database, session=client_session)
        assert pymongo_api.get_detail(db=pymongo_database, session=client_session)

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, None, None, None, {})],
        indirect=['pymongo_api'],
    )
    def test_get_all(self, pymongo_api: SyncAPI, pymongo_database, client_session):
        """Test get_all method."""
        response = pymongo_api.get_all(db=pymongo_database, session=client_session)
        assert len(list(response)) == NUMBER_OF_OBJECTS
        response = pymongo_api.get_all(db=pymongo_database, session=client_session, limit=10)
        assert len(list(response)) == 10
        response = pymongo_api.get_all(
            db=pymongo_database,
            session=client_session,
            limit=5,
            skip=5,
        )
        response_list = list(response)
        assert response_list[0]['id'] == 6  # first 5 result skipped
        assert len(response_list) == 5  # limit value

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', 'id', None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_get_detail(self, pymongo_api: SyncAPI, pymongo_database):
        """Test get_detail method with id as pk_field."""
        response = pymongo_api.get_detail(
            db=pymongo_database,
            field_value=1,
        )
        assert response['id'] == 1

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_get_detail_pk_field(self, pymongo_api: SyncAPI, pymongo_database):
        """Test get_detail method with None as pk_field."""
        with pytest.raises(AttributeError):
            pymongo_api.get_detail(
                db=pymongo_database,
                field_value=1,
            )

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, None, None, None, {'id': 15})],
        indirect=['pymongo_api'],
    )
    def test_get_detail_statement(self, pymongo_api: SyncAPI, pymongo_database):
        """Test get_detail method with None as pk_field and custom statement."""
        response = pymongo_api.get_detail(
            db=pymongo_database,
        )
        assert response['id'] == 15

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', 'id', None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_get_detail_not_found(self, pymongo_api: SyncAPI, pymongo_database):
        """Test get_detail method not found."""
        with pytest.raises(HTTPException):
            pymongo_api.get_detail(
                db=pymongo_database,
                field_value=51,
            )

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', 'id', None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_create(self, pymongo_api: SyncAPI, pymongo_database):
        """Test create method."""
        new_doc = {'username': 'john', 'age': 22, 'id': 111}
        pymongo_api.create(
            db=pymongo_database,
            document=new_doc,
        )
        latest = pymongo_api.get_detail(
            db=pymongo_database,
            field_value=new_doc['id'],
        )
        assert latest['username'] == 'john'
        assert latest['age'] == 22

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', 'id', None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_create_invalid_document(self, pymongo_api: SyncAPI, pymongo_database):
        """Test create method invalid document."""
        new_doc = False
        with pytest.raises(TypeError):
            pymongo_api.create(
                db=pymongo_database,
                document=new_doc,
            )

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', 'id', None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_update_one_doc_none(self, pymongo_api: SyncAPI, pymongo_database):
        """Test update_one method document none."""
        first = pymongo_api.get_detail(
            db=pymongo_database,
            field_value=1,
        )
        first_username = first['username']
        assert first_username != 'new_username'
        # update first
        pymongo_api.update_one(
            db=pymongo_database,
            field_value=1,
            update_doc={'username': 'new_username'}
        )
        first_updated = pymongo_api.get_detail(
            db=pymongo_database,
            field_value=1,
        )
        assert first_updated['username'] != first_username
        assert first_updated['username'] == 'new_username'

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, None, None, None, {'id': 20})],
        indirect=['pymongo_api'],
    )
    def test_update_one_statement_not_none(
            self,
            pymongo_api: SyncAPI,
            pymongo_database,
    ):
        """Test update_one method statement not none."""
        first = pymongo_api.get_detail(
            db=pymongo_database,
        )
        first_username = first['username']
        assert first_username != 'new_username'
        # update first
        pymongo_api.update_one(
            db=pymongo_database,
            update_doc={'username': 'new_username'}
        )
        first_updated = pymongo_api.get_detail(
            db=pymongo_database,
        )
        assert first_updated['username'] != first_username
        assert first_updated['username'] == 'new_username'

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', 'id', None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_update_one_document_not_none(
            self,
            pymongo_api: SyncAPI,
            pymongo_database,
    ):
        """Test update_one method document not none."""
        first = pymongo_api.get_detail(
            db=pymongo_database,
            field_value=1,
        )
        first_username = first['username']
        assert first_username != 'new_username'

        pymongo_api.pk_field = None
        # update first
        pymongo_api.update_one(
            db=pymongo_database,
            update_doc={'username': 'new_username'},
            document={'id': 1}
        )

        pymongo_api.pk_field = 'id'
        first_updated = pymongo_api.get_detail(
            db=pymongo_database,
            field_value=1,
        )
        assert first_updated['username'] != first_username
        assert first_updated['username'] == 'new_username'

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', 'id', None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_delete(
            self,
            pymongo_api: SyncAPI,
            pymongo_database,
    ):
        """Test delete method."""

        first = pymongo_api.get_detail(
            db=pymongo_database,
            field_value=1,
        )
        assert first['id'] == 1
        pymongo_api.delete(pymongo_database, field_value=1)
        with pytest.raises(HTTPException):  # deleted
            pymongo_api.get_detail(
                db=pymongo_database,
                field_value=1,
            )

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_delete_invalid_pk_field(
            self,
            pymongo_api: SyncAPI,
            pymongo_database,
    ):
        """Test delete method invalid pk_field."""
        with pytest.raises(AttributeError):
            pymongo_api.delete(pymongo_database, field_value=1)

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', [], None, None, None, None)],
        indirect=['pymongo_api'],
    )
    def test_delete_invalid_document(
            self,
            pymongo_api: SyncAPI,
            pymongo_database,
    ):
        """Test delete method invalid document."""
        with pytest.raises(HTTPException):
            pymongo_api.delete(pymongo_database, field_value=1)

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, 20, None, {}, {})],
        indirect=['pymongo_api'],
    )
    def test_limit_offset_pagination_first_page(
            self,
            pymongo_api: SyncAPI,
            pymongo_database,
            request_fixture,
    ):
        """Test limit offset pagination first page. Total objects 50."""
        request_fixture.query_params = {}
        response = pymongo_api.get_all_with_pagination(pymongo_database, request_fixture)
        paginate_by = 20
        results = response['results']
        assert results[0]['id'] == 1
        assert results[19]['id'] == 20
        assert response['count'] == paginate_by
        assert response['total'] == NUMBER_OF_OBJECTS
        assert response['page_limit'] == paginate_by
        assert response['total_pages'] == 2
        assert response['previous_page'] is None
        assert response['next_page'] is not None
        assert response['last_page'] is not None

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, 20, None, {}, {})],
        indirect=['pymongo_api'],
    )
    def test_limit_offset_pagination_second_page(
            self,
            pymongo_api: SyncAPI,
            pymongo_database,
            request_fixture,
    ):
        """Test limit offset pagination second page. Total objects 50."""
        request_fixture.query_params = {
            'page': 1,
        }
        response = pymongo_api.get_all_with_pagination(pymongo_database, request_fixture)
        results = response['results']
        assert results[0]['id'] == 21
        assert results[19]['id'] == 40
        assert response['count'] == 20
        assert response['previous_page'] is not None
        assert response['next_page'] is not None
        assert response['last_page'] is not None

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, 20, None, {}, {})],
        indirect=['pymongo_api'],
    )
    def test_limit_offset_pagination_last_page(
            self,
            pymongo_api: SyncAPI,
            pymongo_database,
            request_fixture,
    ):
        """Test limit offset pagination last page. Total objects 50."""
        request_fixture.query_params = {
            'page': 2,
        }
        response = pymongo_api.get_all_with_pagination(pymongo_database, request_fixture)
        results = response['results']
        assert results[0]['id'] == 41
        assert results[9]['id'] == 50
        assert response['count'] == 10
        assert response['previous_page'] is not None
        assert response['next_page'] is None
        assert response['last_page'] is not None

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, 7, PaginationCursor,
          {
              'model': 'test_collection',
              'ordering': ['id'],
              'cursor_prefixes': ['next__', 'prev__']
          },
          {'id': {'$gt': 25}},
          )],
        indirect=['pymongo_api'],
    )
    def test_cursor_pagination_first_page(
            self,
            pymongo_api: SyncAPI,
            pymongo_database,
            request_fixture,
    ):
        """Test limit offset pagination first page. Total objects 50."""
        request_fixture.query_params = {}
        assert pymongo_api.pagination_strategy == PaginationCursor
        response = pymongo_api.get_all_with_pagination(pymongo_database, request_fixture)
        paginate_by = 7
        results = response['results']
        assert results[0]['id'] == 26  # query was id > 25
        assert results[6]['id'] == 32
        assert response['count'] == paginate_by
        assert response['page_limit'] == paginate_by
        assert response['first_page'] is not None
        assert response['previous_page'] is None
        assert response['next_page'] is not None
        assert 'cursor' in response['next_page']

    @pytest.mark.parametrize(
        'pymongo_api',
        [('test_collection', None, 7, PaginationCursor,
          {
              'model': 'test_collection',
              'ordering': ['-id'],  # descending
              'cursor_prefixes': ['next__', 'prev__']
          },
          {'id': {'$gt': 25}},
          )],
        indirect=['pymongo_api'],
    )
    def test_cursor_pagination_first_page_descending(
            self,
            pymongo_api: SyncAPI,
            pymongo_database,
            request_fixture,
    ):
        """Test limit offset pagination first page descending order. Total objects 50."""
        request_fixture.query_params = {}
        assert pymongo_api.pagination_strategy == PaginationCursor
        response = pymongo_api.get_all_with_pagination(pymongo_database, request_fixture)
        paginate_by = 7
        results = response['results']
        assert results[0]['id'] == 50
        assert results[6]['id'] == 44
        assert response['count'] == paginate_by
        assert response['page_limit'] == paginate_by
        assert response['first_page'] is not None
        assert response['previous_page'] is None
        assert response['next_page'] is not None
        assert 'cursor' in response['next_page']



import abc
import random
import string
from typing import Dict, Type, TypeVar

from pydantic import BaseModel, create_model

from fastapi_views.pagination.annotations import (
    CursorPageParams,
    LimitOffsetPageParams
)

P = TypeVar('P', bound=BaseModel)
S = TypeVar('S', bound=BaseModel)
AnyClassMethod = TypeVar('AnyClassMethod')


class AbstractPage(abc.ABC):
    result_key: str = 'results'

    @property
    @abc.abstractmethod
    def params_cls(self) -> Type[S]:
        """Page params. Expect to return pydantic model cls."""

    def create_page(self, pydantic_model: Type[S]) -> Type[P]:
        """Creates pydantic model Page that is going to be used as response_model in router."""
        field_definitions = self._get_result_field(pydantic_model)
        unique_value = ''.join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(12)
        )
        return create_model(
            f'Page_{unique_value}',
            __validators__=self.set_new_model_validators(),
            __base__=self.params_cls,
            **field_definitions,
        )

    @classmethod
    def include(cls, pydantic_model: Type[S]) -> Type[P]:
        """
        Takes as input pydantic model that will be used as type for result list.
        Returns new pydantic Page object.
        """
        return cls().create_page(pydantic_model)

    def _get_result_field(self, pydantic_model: Type[S]) -> dict[str, tuple[list[Type[S]], None]]:
        """Responsible for setting type of 'list of results' for page."""
        return {
            self.result_key: (list[pydantic_model], None),
        }

    def set_new_model_validators(self, *args, **kwargs) -> Dict[str, AnyClassMethod]:
        """
        Override if there is a need to put some validators to new Page.

        Usage example:

            from pydantic import validator

            def username_alphanumeric(cls, v):
                assert v.isalnum(), 'must be alphanumeric'
                return v

            validators = {
                'username_validator': validator('username')(username_alphanumeric),
                ...
            }
        """


class LimitOffsetPage(AbstractPage):

    @property
    def params_cls(self):
        return LimitOffsetPageParams


class CursorPage(AbstractPage):

    @property
    def params_cls(self):
        return CursorPageParams



from typing import Any, Generic, Mapping, Union

from bson import InvalidDocument
from motor.motor_asyncio import (
    AsyncIOMotorClientSession,
    AsyncIOMotorCursor,
    AsyncIOMotorDatabase
)
from pydantic.dataclasses import dataclass
from pymongo.client_session import ClientSession
from pymongo.cursor import Cursor
from pymongo.database import Database
from pymongo.results import DeleteResult, InsertOneResult, UpdateResult

from fastapi_views.ext.pymongo.core import BaseAPIView
from fastapi_views.settings import BaseConfig
from fastapi_views.types import Record, Table


@dataclass(config=BaseConfig)
class SyncAPI(BaseAPIView, Generic[Table]):
    """
    Sync API with default methods to get, create, update, delete objects.

    Motor types shown in Union's are only provided for readability when called by motor methods.
    """
    model: str = None  # collection name

    def get_all(
            self,
            db: Union[AsyncIOMotorDatabase, Database],
            session: Union[AsyncIOMotorClientSession, ClientSession, None] = None,
            limit: int = 100,
            skip: int = 0,
            **kwargs,
    ) -> Union[AsyncIOMotorCursor, Cursor]:
        """
        Executes mongodb find method.

        Read docs for more information.

            ** Required Arguments **
                - db: motor / pymongo database instance

            ** Optional Arguments **
                - session: motor / pymongo client session instance
                - limit: collection limit
                - skip: collection skip
                - kwargs: additional kwargs passed to find method
        """
        self.validate_session(session)
        self.validate_statement()
        collection = self._get_collection(db)
        return collection.find(self.statement, session=session, **kwargs).limit(limit).skip(skip)

    def _get_detail_statement(self, field_value) -> Mapping[str, Any]:
        if self.statement is not None:
            return self.statement
        self.validate_field_value(field_value)
        if self.pk_field is None:
            self.exceptions.attribute_error('pk_field')
        return {self.pk_field: field_value}

    def get_detail(
            self,
            db: Union[AsyncIOMotorDatabase, Database],
            session: Union[AsyncIOMotorClientSession, ClientSession, None] = None,
            field_value: Any = None,
            **kwargs,
    ) -> Record:
        """
        Executes mongodb findOne method.

        Be aware in case of duplicated document, only first matched will be returned.

            ** Required Arguments **
                - db: motor / pymongo database instance

            ** Optional Arguments **
                - session: motor / pymongo client session instance
                - field_value: should be a unique value that will be used to create
                                filter document like {self.pk_field: field_value} for
                                getting single object. Used only if custom statement is not
                                provided
                - kwargs: additional kwargs passed to find_one method
        """
        self.validate_session(session)
        statement = self._get_detail_statement(field_value)
        collection = self._get_collection(db)
        result = collection.find_one(statement, session=session, **kwargs)
        if result is None:
            self.not_found_exception(field_value)
        return result

    def create(
            self,
            db: Union[AsyncIOMotorDatabase, Database],
            document: Mapping[str, Any],
            session: Union[AsyncIOMotorClientSession, ClientSession, None] = None,
            **kwargs,
    ) -> InsertOneResult:
        """
        Executes mongodb insertOne method.

            ** Required Arguments **
                - db: motor / pymongo database instance
                - document: json object from which is created document in database

            ** Optional Arguments **
                - session: motor / pymongo client session instance
                - kwargs: additional kwargs passed to insert_one method
        """
        self.validate_session(session)
        if not isinstance(document, dict):
            raise TypeError('Document is expected to be type of dict')

        return self._get_collection(db).insert_one(document, session=session, **kwargs)

    def delete(
            self,
            db: Union[AsyncIOMotorDatabase, Database],
            field_value: Any,
            session: Union[AsyncIOMotorClientSession, ClientSession, None] = None,
            **kwargs,
    ) -> DeleteResult:
        """
        Executes mongodb deleteOne method.

            ** Required Arguments **
                - db: motor / pymongo database instance
                - field_value: should be a unique value that will be used to create
                                filter document like {self.pk_field: field_value} to delete object

            ** Optional Arguments **
                - session: motor / pymongo client session instance
                - kwargs: additional kwargs passed to insert_one method
        """
        self.validate_session(session)
        self.validate_field_value(field_value)

        collection = self._get_collection(db)
        try:
            if self.pk_field is None:
                self.exceptions.attribute_error('pk_field')
            return collection.delete_one({self.pk_field: field_value}, session=session, **kwargs)
        except (TypeError, InvalidDocument) as e:
            self.invalid_document(str(e))

    def _get_collection(self, db):
        return getattr(db, self.model)

    def _create_update_doc(
            self,
            field_value: Any = None,
            document: Mapping[str, Any] = None,
    ):
        if self.statement is not None:
            return self.statement
        if self.pk_field is None and document is None:
            raise AttributeError('pk_field or document has to be provided.')
        if self.pk_field is None or not self.pk_field:
            self.exceptions.attribute_error('pk_field')
        self.validate_field_value(field_value)
        return {self.pk_field: field_value}

    def update_one(
            self,
            db: Union[AsyncIOMotorDatabase, Database],
            update_doc: Mapping[str, Any],
            field_value: Any = None,
            session: Union[AsyncIOMotorClientSession, ClientSession, None] = None,
            document: Mapping[str, Any] = None,
            **kwargs,
    ) -> UpdateResult:
        """
        Executes mongodb updateOne method.

        When get_statement is given then statement's returned object is used as filter document,
        otherwise you can pass your own filter document with document argument. Last option is
        just to set pk_field and send field_value

            ** Required Arguments **
                - db: motor / pymongo database instance
                - update_doc: document with data to be updated by $set

            ** Optional Arguments **
                - session: motor / pymongo client session instance
                - document: json object used as a filter to find an object to update
                - field_value: should be a unique value that will be used to create
                                filter document like {self.pk_field: field_value} for
                                updating single object
                - kwargs: additional kwargs passed to find_one method
        """
        self.validate_session(session)
        if document is None:
            document = self._create_update_doc(field_value, document)

        collection = self._get_collection(db)
        return collection.update_one(document, {"$set": update_doc}, session=session, **kwargs)



import abc
import math
from dataclasses import field
from functools import cached_property
from typing import Any, Mapping, Type, Union

from fastapi import HTTPException, status
from fastapi.datastructures import QueryParams
from fastapi.requests import Request
from pydantic.dataclasses import dataclass

from fastapi_views.pagination.annotations import (
    CursorResponse,
    LimitOffsetResponse,
    PaginationKwargs,
    Table
)
from fastapi_views.settings import BaseConfig
from fastapi_views.types import Record
from fastapi_views.utils import (
    decode_value,
    dict_to_query_string,
    encode_value,
    get_regex_group
)


class PaginationBase(abc.ABC):

    @abc.abstractmethod
    def paginate_object_list(self, *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def get_paginated_response(self, object_list):
        raise NotImplementedError


class LimitPageException:
    def query_type_exception(self, param: str, expected_type: str) -> Exception:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f'Invalid query type, <:{param}> should be type of {expected_type}.',
        )

    def query_int_value_exception(self, param: str, condition: str) -> Exception:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f'Invalid query value, <:{param}> should be {condition}.',
        )


class LimitMixin(LimitPageException):
    paginate_by: int

    def _get_limit(self, query_params: QueryParams) -> int:
        param = 'limit'
        try:
            limit = int(query_params.get('limit', self.paginate_by))
        except ValueError:
            self.query_type_exception(param, 'int')
        else:
            if limit <= 0:
                self.query_int_value_exception(param, 'higher than 0')
            return limit


class PageMixin(LimitPageException):

    def _get_page(self, query_params: QueryParams) -> int:
        param = 'page'
        try:
            page = int(query_params.get(param, 0))
        except ValueError:
            self.query_type_exception(param, 'int')
        else:
            if page < 0:
                self.query_int_value_exception(param, 'higher or equal 0')
            return page


class LimitPageMixin(LimitMixin, PageMixin):
    ...


class BasePaginationCursor(PaginationBase, LimitMixin, abc.ABC):
    """
    Pagination of Ordered Queries.

    Requires model and ordering to be sent via kwargs when calling paginate_api function, like {
            'model': sqlalchemy Model = modelName,
            'ordering': list[str] = ['-id'],
        }
    Negative '-' means descending order when objects are loaded.
    """

    request: Request
    paginate_by: int
    limit: int
    kwargs: Mapping[Any, Any]
    cursor: str
    model_ordering: list[Any]
    object_list: list[Record] = None
    reverse: bool = False
    _objects = None

    def __init__(self):
        self.lt = 'lt'
        self.gt = 'gt'
        self.next_prefix = 'next__'
        self.previous_prefix = 'previous__'

    @abc.abstractmethod
    def _get_query_result(self) -> list[Record]:
        """Execute db query that will return list of records."""

    def _create_model_ordering(self) -> None:
        """
        Override to set behaviour how query will be sorted.

        Ex. self.model_ordering = [('field_one', 1), ('field_two', -1)]
        """
        self.model_ordering = []

    @cached_property
    def model(self) -> Type[Table]:
        return self.kwargs.get('model')

    @property
    def first_order_field(self) -> str:
        return self.kwargs['ordering'][0]

    @cached_property
    def decoded_cursor(self) -> str:
        if self.cursor:
            return decode_value(self.cursor)
        raise ValueError(f'Can not decode {self.cursor}')

    def _setup_next_page_filter_data(self) -> (str, str):
        op = self.lt if self.first_order_field.startswith('-') else self.gt
        return self.next_prefix, op

    def _update_first_ordering_field(self, value: str) -> None:
        self.kwargs['ordering'][0] = value

    def _setup_previous_page_filter_data(self) -> (str, str):
        if self.first_order_field.startswith('-'):  # start point desc
            value = get_regex_group('(?<=-).*', self.first_order_field)
            op = self.gt
        else:  # start point asc
            value = f'-{self.first_order_field}'
            op = self.lt

        self._update_first_ordering_field(value)
        self._create_model_ordering()
        self.reverse = True
        return self.previous_prefix, op

    def _get_prefix_and_operator(self, is_next_page: bool) -> (str, str):
        if is_next_page:
            return self._setup_next_page_filter_data()
        return self._setup_previous_page_filter_data()

    def get_objects(self, result: list[Record]) -> list[Record]:
        if self.reverse:
            return list(reversed(result))
        return result

    def get_first_order_field(self) -> str:
        if self.first_order_field.startswith('-'):
            return get_regex_group('(?<=-).*', self.first_order_field)
        return self.first_order_field

    @property
    def count(self) -> int:
        # count default
        if self.object_list is not None:
            return len(self.object_list)
        return 0

    @cached_property
    def additional_count(self) -> int:
        # count of limit + 1
        return len(self._objects)

    @property
    def has_next(self) -> bool:
        # at least one additional result exists.
        if self.cursor is not None and self.get_current_decoded_page == 0:
            return True
        return self.additional_count > self.limit

    @property
    def has_previous(self) -> bool:
        if self.cursor is None:
            return False
        return self.get_current_decoded_page > 0

    @cached_property
    def get_current_decoded_page(self) -> int:
        return int(get_regex_group('(?<=page__).*', self.decoded_cursor))

    @property
    def _get_next_page(self) -> int:
        if self.cursor is None:
            return 1
        return self.get_current_decoded_page + 1

    @property
    def _get_previous_page(self) -> int:
        if self.cursor is None:
            return 0
        return self.get_current_decoded_page - 1

    def _get_next_url(self) -> Union[str, None]:
        if self.has_next:
            last_object = self.object_list[-1]
            if type(last_object) == dict:
                value = last_object.get(self.get_first_order_field())
            else:
                value = getattr(last_object, self.get_first_order_field())
            return encode_value(
                f'{self.next_prefix}{value}'
                f'&page__{self._get_next_page}')
        return None

    def _get_previous_url(self) -> Union[str, None]:
        if self.has_previous:
            first_object = self.object_list[0]
            if type(first_object) == dict:
                value = first_object.get(self.get_first_order_field())
            else:
                value = getattr(first_object, self.get_first_order_field())
            return encode_value(
                f'{self.previous_prefix}{value}'
                f'&page__{self._get_previous_page}')
        return None

    def _build_uri(self, encoded_value=None) -> Union[str, None]:
        regex_result = get_regex_group('^[^?]*', str(self.request.url))
        if encoded_value is None:
            return
        new_query_params = {**self.request.query_params, 'cursor': encoded_value.decode("utf-8")}

        return f'{regex_result}?{dict_to_query_string(new_query_params)}'

    def _get_first_url(self) -> str:
        regex_result = get_regex_group('^[^?]*', str(self.request.url))
        q = {**self.request.query_params, 'limit': self.limit}
        q.pop('cursor', None)
        return f'{regex_result}?{dict_to_query_string(q)}'

    @staticmethod
    def is_ordering_model_field_negative(f_name) -> bool:
        return f_name.startswith('-')

    def _get_object_list(self) -> list[Record]:
        """Final result that will be returned as object list."""
        if self.additional_count <= self.limit:
            return self._objects
        if not self.reverse:
            return self._objects[:self.additional_count - 1]
        return self._objects[1:]

    def _prefixes_init(self):
        if 'cursor_prefixes' in self.kwargs:
            self.next_prefix, self.previous_prefix = self.kwargs['cursor_prefixes']

    def paginate_object_list(
            self,
            request,
            *args,
            **kwargs: PaginationKwargs,
    ) -> list[Record]:
        self.request = request
        self.paginate_by = request.state.paginate_by
        query_params = self.request.query_params

        self.limit = self._get_limit(query_params)
        self.kwargs = kwargs
        self._prefixes_init()
        self.cursor = query_params.get('cursor')

        self._create_model_ordering()
        self._objects = self.get_objects(self._get_query_result())
        self.object_list = self._get_object_list()

        return self.object_list

    def get_paginated_response(self, object_list) -> CursorResponse:
        return {
            'results': object_list,
            'count': self.count,
            'page_limit': self.limit,
            'first_page': self._get_first_url(),
            'next_page': self._build_uri(self._get_next_url()),
            'previous_page': self._build_uri(self._get_previous_url()),
        }


class BasePaginationLimitOffset(PaginationBase, LimitPageMixin, abc.ABC):
    """
    Returns list of objects based on request's query parameters [ limit and page ].

    Big page number slows getting query result due to how OFFSET works in sql.
    The rows skipped by an OFFSET clause still have to be computed inside the server,
    therefore a large OFFSET might be inefficient.
    """

    page: int
    limit: int
    request: Request
    paginate_by: int
    _count = None

    @abc.abstractmethod
    def _get_objects(self) -> list[Record]:
        """Get list of objects from db."""

    @abc.abstractmethod
    @cached_property
    def count(self) -> int:
        """Count objects from query."""

    @cached_property
    def total_pages(self) -> int:
        if self._count == 0:
            return 0
        return math.ceil(self._count / self.limit) - 1

    @property
    def _get_next_url(self) -> Union[str, None]:
        if not self.page < self.total_pages:
            return None
        return self._build_uri(self.page + 1)

    @property
    def _get_previous_url(self) -> Union[str, None]:
        if not self.page > 0:
            return None
        return self._build_uri(self.page - 1)

    @property
    def _get_last_url(self) -> str:
        return self._build_uri(self.total_pages)

    def _build_uri(self, new_page: int) -> str:
        # catch all until first "?" occurrence
        regex_result = get_regex_group('^[^?]*', str(self.request.url))
        new_query_params = {**self.request.query_params, 'limit': self.limit, 'page': new_page}

        return f'{regex_result}?{dict_to_query_string(new_query_params)}'

    def paginate_object_list(
            self,
            request: Request,
            *args,
            **kwargs,
    ) -> list[Record]:
        self.request = request
        self.paginate_by = request.state.paginate_by
        self._count = self.count

        self.page = self._get_page(request.query_params)
        self.limit = self._get_limit(request.query_params)

        return self._get_objects()

    def get_paginated_response(self, object_list: list[Record]) -> LimitOffsetResponse:
        return {
            'results': object_list,
            'count': len(object_list),
            'total': self._count,
            'total_pages': self.total_pages,
            'page_limit': self.limit,
            'next_page': self._get_next_url,
            'previous_page': self._get_previous_url,
            'last_page': self._get_last_url,
        }


@dataclass(config=BaseConfig)
class BasePaginatorAPI(abc.ABC):
    """
    Paginator API. Initial setup.

    Arguments:
        ** request **: fastapi Request object.
        ** strategy **: pagination strategy object
           as default: PaginationLimitOffset / AsyncPaginationLimitOffset
        ** paginate_by **: number of objects per page.
        ** kwargs **: optional params passed to strategy'/s paginate_object_list method
    """
    request: Request
    strategy: Type[PaginationBase] = None
    paginate_by: int = 100
    kwargs: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.strategy_instance: PaginationBase = self.strategy()

        if self.kwargs is None or not self.kwargs or not isinstance(self.kwargs, dict):
            self.kwargs = {
                'model': None,
                'ordering': None,
                'cursor_prefixes': None
            }

        self.request.state.paginate_by = self.paginate_by

    @abc.abstractmethod
    def _strategy_data(self) -> list[Any]:
        """Returns only list of current page objects data."""

    def strategy_response(self) -> dict[str, Any]:
        """Create and return response."""
        return self.strategy_instance.get_paginated_response(self._strategy_data)



from typing import Optional, Type, TypedDict, TypeVar, Union

from fastapi.requests import Request
from pydantic import BaseModel
from sqlalchemy.orm import Session

Table = TypeVar("Table")
Record = TypeVar('Record')


class LimitOffsetPageParams(BaseModel):
    count: Union[int, None]
    total: Union[int, None]
    total_pages: Union[int, None]
    page_limit: Union[int, None]
    next_page: Union[str, None]
    previous_page: Union[str, None]
    last_page: Union[str, None]


class CursorPageParams(BaseModel):
    count: Union[int, None]
    page_limit: Union[int, None]
    first_page: Union[str, None]
    next_page: Union[str, None]
    previous_page: Union[str, None]


class PaginationKwargs(TypedDict):
    model: Union[Type[Table], None]
    ordering: Union[list[str], None]
    cursor_prefixes: Union[list[str], None]


class LimitOffsetResponse(TypedDict):
    results: list[Record]
    count: int
    total: int
    total_pages: int
    page_limit: int
    next_page: Optional[str]
    previous_page: Optional[str]
    last_page: str


class CursorResponse(TypedDict):
    results: list[Record]
    count: int
    page_limit: int
    first_page: str
    next_page: Optional[str]
    previous_page: Optional[str]



from unittest.mock import MagicMock

import pytest
import pytest_asyncio
from motor.motor_asyncio import AsyncIOMotorClientSession
from pymongo.client_session import ClientSession

from fastapi_views.tests._db_sessions import MotorDB, PymongoDB


@pytest.fixture
def pymongo_database():
    _db = PymongoDB()
    _db.create_collection('test_collection')
    _db.create_documents()

    yield _db.db


@pytest_asyncio.fixture
async def motor_collection():
    _db = MotorDB()
    await _db.create_documents()

    yield _db.collection


@pytest_asyncio.fixture
async def motor_database():
    _db = MotorDB()
    await _db.create_documents()

    yield _db.db


@pytest.fixture()
def client_session():
    class SessionMock:
        pass

    return MagicMock(spec=ClientSession, return_value=SessionMock())


@pytest.fixture()
def motor_client_session():
    class SessionMock:
        pass

    return MagicMock(spec=AsyncIOMotorClientSession, return_value=SessionMock())



"""
Since most of the tests are common for sync and async,
there will be only tested repository method calls.
"""

import pytest
import pytest_asyncio
from fastapi import HTTPException
from sqlalchemy import select

from fastapi_views.ext.sql_alchemy.async_api import AsyncAPI
from fastapi_views.ext.sql_alchemy.pagination.core import AsyncPaginationCursor
from fastapi_views.tests._db_sessions import (
    AsyncDatabaseSession,
    DatabaseSession,
    DatabaseSessionSameDateTime
)
from fastapi_views.tests._tmp_db import SimpleTestOrm

NUMBER_OF_OBJECTS = 15


@pytest_asyncio.fixture
async def async_session():
    async_db_session = AsyncDatabaseSession()
    await async_db_session.init()
    await async_db_session.create_all()
    yield async_db_session.session
    await async_db_session.session.close()
    await async_db_session.delete_all()


@pytest.fixture
def sync_session():
    db_session = DatabaseSession()
    db_session.init()
    db_session.create_all()
    yield db_session.session
    db_session.close_and_drop()


@pytest_asyncio.fixture
async def session_same_datetime():
    db_session = DatabaseSessionSameDateTime()
    await db_session.init()
    await db_session.create_all()
    yield db_session.session
    await db_session.session.close()
    await db_session.delete_all()


@pytest.fixture
def api(request):
    model, pk_field, statement = request.param
    return AsyncAPI(
        model=model,
        pk_field=pk_field,
        statement=statement,
        paginate_by=3,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'api',
    [(SimpleTestOrm, None, select(SimpleTestOrm))],
    indirect=['api'],
)
async def test_get_all(async_session, api):
    """Test if returns all created objects."""
    assert len(await api.get_all(session=async_session)) == NUMBER_OF_OBJECTS
    assert await api._count(async_session, select(SimpleTestOrm)) == NUMBER_OF_OBJECTS


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'api',
    [(SimpleTestOrm, None, select(SimpleTestOrm))],
    indirect=['api'],
)
async def test_get_all_with_pagination_async_session(async_session, request_fixture, api):
    """Test get_all limit, offset pagination using async session."""
    request_fixture.query_params = {}
    response = await api.get_all_with_pagination(session=async_session, request=request_fixture)
    # async session creates 15 objects
    assert response['total'] == 15
    assert response['count'] == 3
    assert response['page_limit'] == 3
    assert 'results' in response
    assert 'next_page' in response
    assert 'previous_page' in response
    assert 'last_page' in response

    assert type(response['results']) == list
    assert len(response['results']) == response['count']


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'api',
    [(SimpleTestOrm, None, select(SimpleTestOrm))],
    indirect=['api'],
)
async def test_get_all_with_pagination_sync_session(sync_session, request_fixture, api):
    """Test get_all limit, offset pagination using sync session."""
    with pytest.raises(TypeError):
        await api.get_all_with_pagination(session=sync_session, request=request_fixture)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'api',
    [(SimpleTestOrm, None, select(SimpleTestOrm).where(SimpleTestOrm.id > 43))],
    indirect=['api'],
)
async def test_get_all_with_pagination_cursor(session_same_datetime, request_fixture, api):
    """Test get_all cursor pagination using sync session."""
    request_fixture.query_params = {}
    api.pagination_kwargs = {
        'model': SimpleTestOrm,
        'ordering': ['-time_created'],
        'cursor_prefixes': ['next__', 'prev__']
    }
    api.pagination_strategy = AsyncPaginationCursor
    api.paginate_by = 6
    response = await api.get_all_with_pagination(
        session=session_same_datetime,
        request=request_fixture,
    )
    assert response['results'][0].id == 44


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'api',
    [(SimpleTestOrm, None, select(SimpleTestOrm).where(SimpleTestOrm.id > 43))],
    indirect=['api'],
)
async def test_get_all_with_pagination_cursor_2(session_same_datetime, request_fixture, api):
    """Test get_all cursor pagination using sync session with two ordering fields."""
    request_fixture.query_params = {}
    api.pagination_kwargs = {
        'model': SimpleTestOrm,
        'ordering': ['-time_created', '-id'],
        'cursor_prefixes': ['next__', 'prev__']
    }
    api.pagination_strategy = AsyncPaginationCursor
    api.paginate_by = 6
    response = await api.get_all_with_pagination(
        session=session_same_datetime,
        request=request_fixture,
    )
    assert response['results'][0].id == 49


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'api',
    [(SimpleTestOrm, 'id', None)],
    indirect=['api'],
)
async def test_get_detail(async_session, api):
    """Test get_detail method."""
    result = await api.get_detail(async_session, field_value=2)
    assert result.id == 2
    assert type(result).__name__ == 'SimpleTestOrm'

    api.statement = select(SimpleTestOrm).where(
        SimpleTestOrm.id == 5,
    )
    result = await api.get_detail(async_session)
    assert result.id == 5

    api.statement = select(SimpleTestOrm)
    with pytest.raises(ValueError):
        await api.get_detail(async_session)

    api.statement = None
    with pytest.raises(HTTPException):
        await api.get_detail(
            session=async_session,
            field_value=17,
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'api',
    [(SimpleTestOrm, None, None)],
    indirect=['api'],
)
async def test_create(async_session, api):
    """Test create method."""
    result = await api.create(
        session=async_session
    )
    assert result.id > NUMBER_OF_OBJECTS
    assert type(result).__name__ == 'SimpleTestOrm'

    with pytest.raises(HTTPException):
        await api.create(
            session=async_session,
            data={
                'id': 'invalid'
            }
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'api',
    [(SimpleTestOrm, 'id', None)],
    indirect=['api'],
)
async def test_delete(async_session, api):
    """Test delete method."""
    assert await api._count(async_session, select(SimpleTestOrm)) == NUMBER_OF_OBJECTS
    await api.delete(
        session=async_session,
        field_value=13,
    )
    with pytest.raises(HTTPException):
        await api.get_detail(async_session, 13)

    assert await api._count(async_session, select(SimpleTestOrm)) == NUMBER_OF_OBJECTS - 1


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'api',
    [(SimpleTestOrm, None, None)],
    indirect=['api'],
)
async def test_delete_pk_field_error(async_session, api):
    """Test delete method pk_field error."""
    with pytest.raises(AttributeError):
        await api.delete(
            session=async_session,
            field_value=13,
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'api',
    [(SimpleTestOrm, 'id', None)],
    indirect=['api'],
)
async def test_update_one(async_session, api):
    """Test update_one method."""
    await api.update_one(
        session=async_session,
        field_value=1,
        data={
            'id': 100
        }
    )
    assert await api.get_detail(async_session, 100) is not None
    with pytest.raises(HTTPException):
        await api.get_detail(async_session, 1)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'api',
    [(SimpleTestOrm, 'id', None)],
    indirect=['api'],
)
async def test_update_one_attrs(async_session, api):
    """Test update_one method attrs."""
    with pytest.raises(TypeError):
        await api.update_one()
    with pytest.raises(AttributeError):
        await api.update_one(session=None, field_value=2)

    with pytest.raises(TypeError):
        await api.update_one(session='not_valid_type', field_value=2)

    with pytest.raises(HTTPException):
        await api.update_one(session=async_session, field_value=2)

    api.pk_field = None
    with pytest.raises(AttributeError):
        await api.update_one(session=async_session, field_value=2, data={'id': 100})

    api.pk_field = 'id'
    with pytest.raises(HTTPException):
        await api.update_one(session=async_session, field_value=2, data={'id': 'invalid'})



from typing import Any, Generic, Mapping, Union

from fastapi.requests import Request
from motor.motor_asyncio import (
    AsyncIOMotorClientSession,
    AsyncIOMotorCursor,
    AsyncIOMotorDatabase
)
from pydantic.dataclasses import dataclass
from pymongo.results import DeleteResult, InsertOneResult, UpdateResult

from fastapi_views.ext.motor.core import APIErrorHandlers, BaseAPIView
from fastapi_views.ext.motor.pagination.core import async_paginate_api
from fastapi_views.ext.pymongo.sync_api import SyncAPI as PymongoAPI
from fastapi_views.settings import BaseConfig
from fastapi_views.types import Record, Table


@dataclass(config=BaseConfig)
class AsyncAPI(APIErrorHandlers, PymongoAPI, BaseAPIView, Generic[Table]):
    """Async API with default methods to get, create, update, delete objects."""

    model: str = None

    async def get_all_with_pagination(
            self,
            db: AsyncIOMotorDatabase,
            request: Request,
            session: AsyncIOMotorClientSession = None,
            **kwargs,
    ):
        """Returns multiple objects using pagination."""
        self.validate_session(session)
        self.validate_pagination(request)
        self.pagination_kwargs['collection'] = getattr(db, self.model)
        self.pagination_kwargs['statement_kwargs'] = kwargs

        paginate_response = await async_paginate_api(
            statement=self.statement,
            request=request,
            session=session,
            paginate_by=self.paginate_by,
            strategy=self.pagination_strategy,
            kwargs=self.pagination_kwargs,
        )
        return paginate_response

    def get_all(
            self,
            db: AsyncIOMotorDatabase,
            session: Union[AsyncIOMotorClientSession, None] = None,
            limit: int = 100,
            skip: int = 0,
            **kwargs,
    ) -> AsyncIOMotorCursor:
        """
        Executes mongodb find method.

        Read docs for more information.

        https://motor.readthedocs.io/en/stable/tutorial-asyncio.html#querying-for-more-than-one-document

            ** Required Arguments **
                - db: motor database instance

            ** Optional Arguments **
                - session: motor client session instance
                - limit: collection limit
                - skip: collection skip
                - kwargs: additional kwargs passed to find method
        """
        return super().get_all(db, session, limit, skip, **kwargs)

    async def get_detail(
            self,
            db: AsyncIOMotorDatabase,
            session: Union[AsyncIOMotorClientSession, None] = None,
            field_value: Any = None,
            **kwargs,
    ) -> Record:
        """
        Executes mongodb findOne method.

        Be aware in case of duplicated document, only first matched will be returned.

            ** Required Arguments **
                - db: motor database instance

            ** Optional Arguments **
                - session: motor client session instance
                - field_value: should be a unique value that will be used to create
                                filter document like {self.pk_field: field_value} for
                                getting single object. Used only if custom statement is not
                                provided
                - kwargs: additional kwargs passed to find_one method
        """
        return await super().get_detail(db, session, field_value, **kwargs)

    async def create(
            self,
            db: AsyncIOMotorDatabase,
            document: Mapping[str, Any],
            session: Union[AsyncIOMotorClientSession, None] = None,
            **kwargs,
    ) -> InsertOneResult:
        """
        Executes mongodb insertOne method.

            ** Required Arguments **
                - db: motor database instance
                - document: json object from which is created document in database

            ** Optional Arguments **
                - session: motor client session instance
                - kwargs: additional kwargs passed to insert_one method
        """
        return await super().create(db, document, session, **kwargs)

    async def delete(
            self,
            db: AsyncIOMotorDatabase,
            field_value: Any,
            session: AsyncIOMotorClientSession = None,
            **kwargs,
    ) -> DeleteResult:
        """
        Executes mongodb deleteOne method.

            ** Required Arguments **
                - db: motor database instance
                - field_value: should be a unique value that will be used to create
                                filter document like {self.pk_field: field_value} to delete object

            ** Optional Arguments **
                - session: motor client session instance
                - kwargs: additional kwargs passed to insert_one method
        """
        return await super().delete(db, field_value, session, **kwargs)

    async def update_one(
            self,
            db: AsyncIOMotorDatabase,
            update_doc: Mapping[str, Any],
            field_value: Any = None,
            session: AsyncIOMotorClientSession = None,
            document: Mapping[str, Any] = None,
            **kwargs,
    ) -> UpdateResult:
        """
        Executes mongodb updateOne method.

        When get_statement is given then statement's returned object is used as filter document,
        otherwise you can pass your own filter document with document argument. Last option is
        just to set pk_field and send field_value

            ** Required Arguments **
                - db: motor database instance
                - update_doc: document with data to be updated by $set

            ** Optional Arguments **
                - session: motor client session instance
                - document: json object used as a filter to find an object to update
                - field_value: should be a unique value that will be used to create
                                filter document like {self.pk_field: field_value} for
                                updating single object
                - kwargs: additional kwargs passed to find_one method
        """
        return await super().update_one(db, update_doc, field_value, session, document, **kwargs)



import base64
import re
from binascii import Error
from typing import Union

from fastapi import HTTPException, status
from sqlalchemy.engine import Row


def scalars(response: Union[dict, list], data_key: str = 'results'):
    """
    Get object list from response and return first value from each row.

    ex. dict input
        {
            '[data_key]': [(sqlAlchemyRowObject, ), (sqlAlchemyRowObject, )],
            'total_pages': 10,
            'count':2,
            ...
        }

        output
        {
            '[data_key]': [sqlAlchemyRowObject, sqlAlchemyRowObject],
            'total_pages': 10,
            'count':2,
            ...
        }

    ex. list input
        [(sqlAlchemyRowObject, ), (sqlAlchemyRowObject, )]

        output
        [sqlAlchemyRowObject, sqlAlchemyRowObject]
    """
    if isinstance(response, dict):
        object_list: list[Row] = response.get(data_key, [])
        if object_list and any(isinstance(row, Row) for row in object_list):
            response[data_key] = [row[0] for row in object_list]
    elif isinstance(response, list):
        response = [row[0] for row in response]
    return response


def get_regex_group(regex: str, search_in: str):
    url_regex = re.compile(f"{regex}")
    mo = url_regex.search(search_in)
    if mo is not None:
        return mo.group()
    raise ValueError('No result found')


def encode_value(value):
    return base64.b64encode(bytes(str(value), "utf-8"))


def decode_value(value):
    try:
        base64.b64decode(value.encode("utf-8")).decode("utf-8", "ignore")
    except Error:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail='Incorrect cursor')
    return base64.b64decode(value.encode("utf-8")).decode("utf-8", "ignore")


def dict_to_query_string(new_query_params) -> str:
    new_string = ''
    for index, (k, v) in enumerate(new_query_params.items()):
        new_string = f'{new_string}{k}={v}'
        if len(new_query_params) - 1 != index:
            new_string = f'{new_string}&'
    return new_string



import abc
from typing import Any, Generic, Type, Union

from fastapi import HTTPException, status
from fastapi.requests import Request
from pydantic import validator
from pydantic.dataclasses import dataclass

from fastapi_views.pagination.core import BasePaginationCursor, PaginationBase
from fastapi_views.settings import BaseConfig
from fastapi_views.types import Table


class RepositoryExceptions:

    def __init__(self, model: Type[Table]):
        self.model = model

    def attribute_error(self, field_name: str) -> None:
        raise AttributeError(f'{field_name} is required.')

    def request_type_error(self) -> None:
        raise TypeError('Request must be type of fastapi.requests Request.')

    def validate_request(self, request) -> None:
        if request is None:
            self.attribute_error('request')
        if not isinstance(request, Request):
            self.request_type_error()

    def not_found_exception(self, unique: Any):
        """Object not found exception."""
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f'{self.model.__name__} with <{type(unique)}:{unique}> does not exist.'
        )

    @staticmethod
    def integrity_error(error):
        """Integrity exception."""
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=error,
        )

    @staticmethod
    def empty_update_error():
        """Integrity exception."""
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Nothing to update.',
        )


@dataclass(config=BaseConfig)
class BaseRepositoryView(Generic[Table], abc.ABC):
    """Base repository API class."""

    model: Type[Table] = None
    paginate_by: int = None
    pagination_strategy: Type[PaginationBase] = None
    pagination_kwargs: Union[dict[str, Any], None] = None

    def __post_init__(self):
        self.exceptions = RepositoryExceptions(model=self.model)

    @validator('model')
    def model_can_not_be_empty(cls, value: Table):
        """Model field validator."""
        if not value or value is None:
            raise ValueError('Model field is required.')
        return value

    def validate_pagination(self, request):
        if self.pagination_strategy and issubclass(self.pagination_strategy, BasePaginationCursor):
            if not self.pagination_kwargs['ordering']:
                raise ValueError('At least one ordering field is required.')
        self.exceptions.validate_request(request)
        if self.paginate_by is None:
            self.exceptions.attribute_error('paginate_by')
        if not isinstance(self.paginate_by, int) or isinstance(self.paginate_by, bool):
            raise TypeError('paginate_by must be type of int')
        if self.paginate_by <= 0:
            raise ValueError('paginate_by value should be higher than 0')

    @abc.abstractmethod
    def get_all(self, *args, **kwargs):
        """Returns multiple objects."""

    @abc.abstractmethod
    def get_detail(self, *args, **kwargs):
        """Return single object."""

    @abc.abstractmethod
    def create(self, *args, **kwargs):
        """Perform object creation."""

    @abc.abstractmethod
    def delete(self, *args, **kwargs):
        """Remove object."""

    @abc.abstractmethod
    def update_one(self, *args, **kwargs):
        """Perform single object update."""



from unittest.mock import patch

import pytest
from sqlalchemy import select

from fastapi_views.ext.sql_alchemy.pagination.core import (
    PaginationCursor,
    PaginationLimitOffset,
    PaginatorAPI,
    paginate_api
)
from fastapi_views.pagination.annotations import CursorResponse
from fastapi_views.tests._db_sessions import DatabaseSession
from fastapi_views.tests._tmp_db import SimpleTestOrm
from fastapi_views.utils import encode_value


@pytest.fixture
def session():
    db_session = DatabaseSession(number=10)
    db_session.init()
    db_session.create_all()
    yield db_session.session
    db_session.close_and_drop()


@pytest.fixture()
def limit_offset_api(request_fixture, request, session):
    request_fixture.query_params = {
        'page': request.param,
        'limit': 3,
    }
    return paginate_api(
        statement=select(SimpleTestOrm),
        request=request_fixture,
        session=session,
    )


@pytest.fixture()
def cursor_api(request_fixture, request, session):
    cursor, ordering = request.param
    request_fixture.query_params = {
        'limit': 2,
        'cursor': cursor,
    }
    return paginate_api(
        statement=select(SimpleTestOrm),
        request=request_fixture,
        session=session,
        strategy=PaginationCursor,
        kwargs={
            'model': SimpleTestOrm,
            'ordering': ordering,
            'cursor_prefixes': ['custom_next__', 'custom_prev__']
        }
    )


class TestPaginatorAPI:

    @patch('fastapi_views.ext.sql_alchemy.pagination.core.PaginatorAPI.strategy_response')
    def test_init(self, response_mock, request_fixture, session):
        """Test if values are initialized as expected."""
        response_mock.return_value = {}

        paginator = PaginatorAPI(
            statement=select(SimpleTestOrm),
            request=request_fixture,
            session=session,
            paginate_by=2
        )
        assert paginator.statement is not None
        assert paginator.request is not None
        assert paginator.session is not None
        assert paginator.strategy is not None
        assert paginator.paginate_by is not None
        assert paginator.strategy == PaginationLimitOffset
        assert paginator.paginate_by == 2
        assert paginator.kwargs == {
            'model': None,
            'ordering': None,
            'cursor_prefixes': None,
        }


class TestLimitOffset:

    @pytest.mark.parametrize('limit_offset_api', [0, 1, 2, 3], indirect=['limit_offset_api'])
    def test_strategy_response_common(self, limit_offset_api: dict):
        """Test pagination response the same common values in every page."""
        assert limit_offset_api['total'] == 10
        assert limit_offset_api['total_pages'] == 3
        assert limit_offset_api['page_limit'] == 3

    @pytest.mark.parametrize('limit_offset_api', [0], indirect=['limit_offset_api'])
    def test_strategy_response(self, limit_offset_api: dict):
        """Test pagination response starting at page 0."""
        assert limit_offset_api['count'] == 3
        assert limit_offset_api['next_page'] == 'https://example.com?page=1&limit=3'
        assert limit_offset_api['previous_page'] is None
        assert limit_offset_api['last_page'] == 'https://example.com?page=3&limit=3'

    @pytest.mark.parametrize('limit_offset_api', [1], indirect=['limit_offset_api'])
    def test_strategy_response_1(self, limit_offset_api: dict):
        """Test pagination response starting at page 1."""
        assert limit_offset_api['count'] == 3
        assert limit_offset_api['next_page'] == 'https://example.com?page=2&limit=3'
        assert limit_offset_api['previous_page'] == 'https://example.com?page=0&limit=3'
        assert limit_offset_api['last_page'] == 'https://example.com?page=3&limit=3'

    @pytest.mark.parametrize('limit_offset_api', [3], indirect=['limit_offset_api'])
    def test_strategy_response_last(self, limit_offset_api: dict):
        """Test pagination response starting at last page."""
        assert limit_offset_api['count'] == 1
        assert limit_offset_api['next_page'] is None
        assert limit_offset_api['previous_page'] == 'https://example.com?page=2&limit=3'
        assert limit_offset_api['last_page'] == 'https://example.com?page=3&limit=3'


cursor_to_load_last_page = encode_value('custom_next__8&page__4').decode("utf-8")
cursor_to_load_last_page_reverse = encode_value('custom_next__3&page__4').decode("utf-8")


def assert_result(result: CursorResponse, next_cursor=None, prev_cursor=None):
    """
    Count is the same for all results because in our test we have 10 objects,
    which paginated by 2 gives equal number of objects per page.
    """
    assert result['count'] == 2
    assert result['page_limit'] == 2
    assert result['first_page'] == 'https://example.com?limit=2'
    if next_cursor is not None:
        next_page_url = f'https://example.com?limit=2&cursor={next_cursor.decode("utf-8")}'
        assert result['next_page'] == next_page_url
    else:
        assert result['next_page'] is None

    if prev_cursor is not None:
        prev_url = f'https://example.com?limit=2&cursor={prev_cursor.decode("utf-8")}'
        assert result['previous_page'] == prev_url
    else:
        assert result['previous_page'] is None


class TestCursorAscendingOrder:
    """We assume in tests that we have 10 objects ordered by id paginated by 2."""

    @staticmethod
    def _get_orm_id(orm):
        return getattr(orm, 'id')

    @staticmethod
    def assert_common(result):
        assert result['count'] == 2
        assert result['page_limit'] == 2
        assert result['first_page'] == 'https://example.com?limit=2'

    @pytest.mark.parametrize('cursor_api', [(None, ['id'])], indirect=['cursor_api'])
    def test_strategy_response(self, cursor_api: dict):
        """Test pagination response."""
        next_page_cursor = encode_value('custom_next__2&page__1')
        assert_result(cursor_api, next_page_cursor)

        api_ids = list(map(self._get_orm_id, cursor_api['results']))
        assert api_ids == [1, 2]

    @pytest.mark.parametrize(
        'cursor_api',
        [('Y3VzdG9tX25leHRfXzImcGFnZV9fMQ==', ['id'])],
        indirect=['cursor_api'],
    )
    def test_strategy_response_with_cursor(self, cursor_api: dict):
        """Test pagination response 'at page 1'."""
        next_page_cursor = encode_value('custom_next__4&page__2')
        previous_page_cursor = encode_value('custom_prev__3&page__0')
        assert_result(cursor_api, next_page_cursor, previous_page_cursor)

        api_ids = list(map(self._get_orm_id, cursor_api['results']))
        assert api_ids == [3, 4]

    @pytest.mark.parametrize(
        'cursor_api', [(cursor_to_load_last_page, ['id'])],
        indirect=['cursor_api'],
    )
    def test_strategy_response_with_cursor_last_page(self, cursor_api: dict):
        """Test pagination response 'in last page'."""
        previous_page_cursor = encode_value('custom_prev__9&page__3')
        assert_result(cursor_api, prev_cursor=previous_page_cursor)

        api_ids = list(map(self._get_orm_id, cursor_api['results']))
        assert api_ids == [9, 10]


class TestCursorDescendingOrder:
    """We assume in tests that we have 10 objects ordered by id desc paginated by 2."""

    @staticmethod
    def _get_orm_id(orm):
        return getattr(orm, 'id')

    @pytest.mark.parametrize('cursor_api', [(None, ['-id'])], indirect=['cursor_api'])
    def test_strategy_response(self, cursor_api: dict):
        """Test pagination response."""
        next_page_cursor = encode_value('custom_next__9&page__1')
        assert_result(cursor_api, next_page_cursor)

        api_ids = list(map(self._get_orm_id, cursor_api['results']))
        assert api_ids == [10, 9]

    @pytest.mark.parametrize(
        'cursor_api',
        [('Y3VzdG9tX25leHRfXzkmcGFnZV9fMQ==', ['-id'])],
        indirect=['cursor_api'],
    )
    def test_strategy_response_with_cursor(self, cursor_api: dict):
        """Test pagination response 'at page 1'."""
        next_page_cursor = encode_value('custom_next__7&page__2')
        previous_page_cursor = encode_value('custom_prev__8&page__0')
        assert_result(cursor_api, next_page_cursor, previous_page_cursor)

        api_ids = list(map(self._get_orm_id, cursor_api['results']))
        assert api_ids == [8, 7]

    @pytest.mark.parametrize(
        'cursor_api',
        [(cursor_to_load_last_page_reverse, ['-id'])],
        indirect=['cursor_api'],
    )
    def test_strategy_response_with_cursor_last_page(self, cursor_api: dict):
        """Test pagination response 'in last page'."""
        previous_page_cursor = encode_value('custom_prev__2&page__3')
        assert_result(cursor_api, prev_cursor=previous_page_cursor)

        api_ids = list(map(self._get_orm_id, cursor_api['results']))
        assert api_ids == [2, 1]



import operator
from functools import cached_property
from typing import Any, Type, Union

from fastapi.requests import Request
from pydantic.dataclasses import dataclass
from sqlalchemy import func, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import InstrumentedAttribute, Session
from sqlalchemy.sql import Select

from fastapi_views.pagination.annotations import (
    CursorResponse,
    LimitOffsetResponse,
    PaginationKwargs
)
from fastapi_views.pagination.core import (
    BasePaginationCursor,
    BasePaginationLimitOffset,
    BasePaginatorAPI,
    PaginationBase
)
from fastapi_views.settings import BaseConfig
from fastapi_views.types import Record
from fastapi_views.utils import get_regex_group


class PaginationCursor(BasePaginationCursor):
    """Cursor pagination."""

    session: Session
    statement: Select

    def _filter_query(self) -> Union[Select, None]:
        decoded_cursor = self.decoded_cursor
        is_next_page = decoded_cursor.startswith(self.next_prefix)
        prefix, op = self._get_prefix_and_operator(is_next_page)

        python_operator = getattr(operator, op)
        regex_result = get_regex_group(f'(?<={prefix})(.*)(?=&page__)', decoded_cursor)
        try:
            regex_result = int(regex_result)
        except ValueError:
            ...
        return self.statement.where(
            python_operator(
                getattr(self.model, self.get_first_order_field()),
                regex_result,
            )
        )

    def _validate_first_ordering_field(self) -> None:
        ordering = self.kwargs.get('ordering')
        first_order = ordering[0]
        first_model_field = self._get_ordering_model_field(
            first_order
        )
        if str(first_model_field.type) in ('BOOLEAN',):
            raise TypeError(f'{first_model_field.type} should can not be used as comparison')

    def _create_model_ordering(self) -> None:
        self._validate_first_ordering_field()
        result = []

        for f_name in self.kwargs.get('ordering'):
            model_field = self._get_ordering_model_field(f_name)

            if self.is_ordering_model_field_negative(f_name):
                model_field = model_field.desc()
            result.append(model_field)
        self.model_ordering = result

    def _get_ordering_model_field(self, f_name: str) -> InstrumentedAttribute:
        model = self.kwargs['model']
        if self.is_ordering_model_field_negative(f_name):
            regex_result = get_regex_group('(?<=-).*', f_name)
            return getattr(model, regex_result)

        return getattr(model, f_name)

    def _get_query_result(self) -> list[Record]:
        statement = self.statement if self.cursor is None else self._filter_query()
        return (self.session.scalars(
            statement.order_by(
                *self.model_ordering,
            ).limit(self.limit + 1))
        ).all()

    def paginate_object_list(
            self,
            request,
            *args,
            **kwargs: PaginationKwargs,
    ) -> list[Record]:
        self.session, self.statement = args
        return super().paginate_object_list(request, **kwargs)

    def get_paginated_response(self, object_list) -> CursorResponse:
        return {
            'results': object_list,
            'count': self.count,
            'page_limit': self.limit,
            'first_page': self._get_first_url(),
            'next_page': self._build_uri(self._get_next_url()),
            'previous_page': self._build_uri(self._get_previous_url()),
        }


class PaginationLimitOffset(BasePaginationLimitOffset):
    """Limit offset pagination."""

    session: Session
    statement: Select

    def _get_objects(self) -> list[Record]:
        return (self.session.execute(
            self.statement.limit(self.limit).offset(self.limit * self.page),
        )).all()

    @cached_property
    def count(self) -> int:
        statement = select(func.count()).select_from(self.statement.subquery())
        return self.session.scalar(statement)

    def paginate_object_list(
            self,
            request: Request,
            *args,
            **kwargs,
    ) -> list[Record]:
        self.session, self.statement = args
        return super().paginate_object_list(request)


@dataclass(config=BaseConfig)
class PaginatorAPI(BasePaginatorAPI):
    """
    Paginator API.

    Arguments:
        ** statement **: sqlalchemy selectable select.
        ** session **: sqlalchemy orm sync session.
    """
    statement: Select = None
    session: Session = None

    def __post_init__(self):
        if self.strategy is None:
            self.strategy = PaginationLimitOffset
        super().__post_init__()

    @property
    def _strategy_data(self) -> list[Any]:
        """Returns only list of current page objects data."""
        return self.strategy_instance.paginate_object_list(
            self.request,
            self.session,
            self.statement,
            **self.kwargs,
        )


def paginate_api(
        statement: Select,
        request: Request,
        session: Session,
        paginate_by: int = 100,
        strategy: Type[PaginationBase] = None,
        kwargs: Union[dict[str, Any], None] = None,
):
    """
    Returns pagination object for given strategy.

    :param statement: sqlalchemy selectable select
    :param request: fastapi Request object.
    :param session: session object
    :param paginate_by: number of objects per page
    :param strategy: pagination strategy object
    :param kwargs: optional params passed to strategy'/s paginate_object_list method
    """
    paginator = PaginatorAPI(
        statement=statement,
        request=request,
        session=session,
        strategy=strategy,
        paginate_by=paginate_by,
        kwargs=kwargs,
    )
    return paginator.strategy_response()


# ----- ASYNC -----

@dataclass(config=BaseConfig)
class AsyncPaginatorAPI(BasePaginatorAPI):
    """Async paginator ext."""
    statement: Select = None
    session: AsyncSession = None

    def __post_init__(self):
        if self.strategy is None:
            self.strategy = AsyncPaginationLimitOffset
        super().__post_init__()

    @property
    async def _strategy_data(self) -> list[Any]:
        """Returns only list of current page objects data."""
        return await self.strategy_instance.paginate_object_list(
            self.request,
            self.session,
            self.statement,
            **self.kwargs,
        )

    async def strategy_response(self) -> dict[str, Any]:
        """Create and return response."""
        return await self.strategy_instance.get_paginated_response(await self._strategy_data)


class AsyncPaginationLimitOffset(BasePaginationLimitOffset):
    """Pagination limit offset async extension."""

    session: AsyncSession
    statement: Select
    _count: int = None

    async def _get_objects(self) -> list[Record]:
        return (await self.session.execute(
            self.statement.limit(self.limit).offset(self.limit * self.page),
        )).all()

    @cached_property
    async def count(self) -> int:
        statement = select(func.count()).select_from(self.statement.subquery())
        return await self.session.scalar(statement)

    async def paginate_object_list(
            self,
            request: Request,
            *args,
            **kwargs,
    ) -> list[Record]:
        self.session, self.statement = args
        self.request = request
        self.paginate_by = request.state.paginate_by
        self._count = await self.count

        self.page = self._get_page(request.query_params)
        self.limit = self._get_limit(request.query_params)

        return await self._get_objects()

    async def get_paginated_response(self, object_list: list[Record]) -> LimitOffsetResponse:
        return super().get_paginated_response(object_list)


class AsyncPaginationCursor(PaginationCursor):
    """Async paginator cursors extension."""

    session: AsyncSession
    statement: Select
    _objects = None

    async def _get_query_result(self) -> list[Record]:
        statement = self.statement if self.cursor is None else self._filter_query()
        return (await self.session.scalars(
            statement.order_by(
                *self.model_ordering,
            ).limit(self.limit + 1))).all()

    async def paginate_object_list(
            self,
            request,
            *args,
            **kwargs: PaginationKwargs,
    ) -> list[Record]:
        self.session, self.statement = args
        self.request = request
        self.paginate_by = request.state.paginate_by
        query_params = self.request.query_params

        self.limit = self._get_limit(query_params)
        self.kwargs = kwargs
        self._prefixes_init()
        self.cursor = query_params.get('cursor')

        self._create_model_ordering()
        self._objects = self.get_objects(await self._get_query_result())
        self.object_list = self._get_object_list()

        return self.object_list

    async def get_paginated_response(self, object_list) -> CursorResponse:
        return super().get_paginated_response(object_list)


async def async_paginate_api(
        statement: Select,
        request: Request,
        session: AsyncSession,
        paginate_by: int = 100,
        strategy: Type[PaginationBase] = None,
        kwargs: Union[dict[str, Any], None] = None,
):
    """
    Returns pagination object for given strategy.

    :param statement: sqlalchemy selectable select
    :param request: fastapi Request object.
    :param session: async session object
    :param paginate_by: number of objects per page
    :param strategy: pagination strategy object
    :param kwargs: optional params passed to strategy'/s paginate_object_list method
    """
    paginator = AsyncPaginatorAPI(
        statement=statement,
        request=request,
        session=session,
        strategy=strategy,
        paginate_by=paginate_by,
        kwargs=kwargs,
    )
    return await paginator.strategy_response()



from functools import cached_property
from typing import Any, Mapping, Type, Union

from fastapi.requests import Request
from pydantic.dataclasses import dataclass
from pymongo import ASCENDING, DESCENDING
from pymongo.client_session import ClientSession
from pymongo.collection import Collection
from pymongo.cursor import Cursor

from fastapi_views.pagination.annotations import PaginationKwargs
from fastapi_views.pagination.core import (
    BasePaginationCursor,
    BasePaginationLimitOffset,
    BasePaginatorAPI,
    PaginationBase
)
from fastapi_views.settings import BaseConfig
from fastapi_views.types import Record
from fastapi_views.utils import get_regex_group


class PaginationCursor(BasePaginationCursor):
    """Cursor pagination."""

    session: ClientSession
    statement: Mapping[str, Any]
    collection: Collection

    def _filter_query(self, cursor: Cursor) -> Cursor:
        """Extends mongodb query by providing additional where clause."""
        is_next_page = self.decoded_cursor.startswith(self.next_prefix)
        prefix, op = self._get_prefix_and_operator(is_next_page)

        regex_result = get_regex_group(f'(?<={prefix})(.*)(?=&page__)', self.decoded_cursor)
        sign = {
            'lt': '<',
            'gt': '>'
        }
        return cursor.where(f'this.{self.get_first_order_field()} {sign.get(op)} {regex_result}')

    def _create_model_ordering(self) -> None:
        """
        Overridden method to set up correct model_ordering attribute.
        Expected list with key, value tuples to assign : [('field', asc/desc), ...]
        """
        result = []
        ordering = self.kwargs.get('ordering')
        if not ordering:
            self.model_ordering = []
        for f_name in ordering:
            if self.is_ordering_model_field_negative(f_name):
                result.append((get_regex_group('(?<=-).*', f_name), DESCENDING))
            else:
                result.append((f_name, ASCENDING))

        self.model_ordering = result

    def _get_query_result(self) -> list[Record]:
        """Final result that will be returned as object list."""
        results = []
        cursor = self.collection.find(self.statement)

        if self.cursor is not None:
            cursor = self._filter_query(cursor)

        if self.model_ordering:
            cursor = cursor.sort(self.model_ordering).limit(self.limit + 1)
        else:
            cursor = cursor.limit(self.limit + 1)
        for document in cursor:
            results.append(document)
        return results

    def paginate_object_list(
            self,
            request,
            *args,
            **kwargs,
    ) -> list[Record]:
        self.session, self.statement, self.collection = args
        return super().paginate_object_list(request, **kwargs)


class PaginationLimitOffset(BasePaginationLimitOffset):
    """Limit offset pagination."""

    session: Union[ClientSession, None]
    statement: Mapping[str, Any]
    collection: Collection

    def _get_objects(self) -> list[Record]:
        """Final result that will be returned as object list."""
        cursor = self.collection.find(
            self.statement,
        ).limit(self.limit).skip(self.limit * self.page)
        results = []
        for document in cursor:
            results.append(document)
        return results

    @cached_property
    def count(self) -> int:
        """Get number of documents for given statement."""
        return self.collection.count_documents(
            self.statement,
            session=self.session,
        )

    def paginate_object_list(
            self,
            request: Request,
            *args,
            **kwargs: PaginationKwargs,
    ) -> list[Record]:
        self.session, self.statement, self.collection = args
        return super().paginate_object_list(request, **kwargs)


@dataclass(config=BaseConfig)
class PaginatorAPI(BasePaginatorAPI):
    """Paginator API."""

    statement: Mapping[str, Any] = None
    session: Union[ClientSession, None] = None

    def __post_init__(self):
        if self.strategy is None:
            self.strategy = PaginationLimitOffset
        super().__post_init__()

    @property
    def _strategy_data(self) -> list[Any]:
        """Returns only list of current page objects data."""
        self.collection = self.kwargs.pop('collection')
        return self.strategy_instance.paginate_object_list(
            self.request,
            self.session,
            self.statement,
            self.collection,
            **self.kwargs,
        )


def paginate_api(
        statement: Mapping[str, Any],
        request: Request,
        session: Union[ClientSession, None] = None,
        paginate_by: int = 100,
        strategy: Type[PaginationBase] = None,
        kwargs: Union[dict[str, Any], None] = None,
):
    """
    Returns pagination object for given strategy.

    :param statement: mongodb filter query
    :param request: fastapi Request object.
    :param session: session object
    :param paginate_by: number of objects per page
    :param strategy: pagination strategy object
    :param kwargs: optional params passed to strategy'/s paginate_object_list method
    """
    paginator = PaginatorAPI(
        statement=statement,
        request=request,
        session=session,
        strategy=strategy,
        paginate_by=paginate_by,
        kwargs=kwargs,
    )
    return paginator.strategy_response()



from typing import Any, Generic, TypeVar

from fastapi import status
from fastapi.requests import Request
from fastapi.responses import Response
from pydantic.dataclasses import dataclass
from sqlalchemy import delete, func, select, update
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.sql import Select

from fastapi_views.ext.sql_alchemy.core import BaseAPIView
from fastapi_views.ext.sql_alchemy.pagination.core import async_paginate_api
from fastapi_views.settings import BaseConfig
from fastapi_views.types import Record

Table = TypeVar("Table")


@dataclass(config=BaseConfig)
class AsyncAPI(BaseAPIView, Generic[Table]):
    """Async API with default methods to get, create, update, delete objects."""

    @staticmethod
    async def _count(session: AsyncSession, statement: Select) -> int:
        """Get number of documents for given statement."""
        count_statement = select(func.count()).select_from(statement.subquery())
        return await session.scalar(count_statement)

    async def get_all(
            self, session: AsyncSession,
            limit: int = 100,
            offset: int = 0,
            **kwargs,
    ) -> list[Record]:
        """Fetch all objects."""
        self.validate_async_session(session)
        self.validate_statement()
        return (await session.execute(
            self.statement.limit(limit).offset(limit * offset),
            **kwargs
        )).all()

    async def get_all_with_pagination(self, request: Request, session: AsyncSession):
        """Returns pagination object."""
        self.validate_async_session(session)
        self.validate_pagination(request)
        paginate_response = await async_paginate_api(
            statement=self.statement,
            request=request,
            session=session,
            paginate_by=self.paginate_by,
            strategy=self.pagination_strategy,
            kwargs=self.pagination_kwargs,
        )
        return paginate_response

    async def get_detail(self, session: AsyncSession, field_value: Any = None, **kwargs) -> Table:
        """Fetch object."""
        self.validate_async_session(session)

        if self.statement is not None:
            statement = self.statement
        else:
            self.validate_field_value(field_value)
            statement = select(self.model).where(getattr(self.model, self.pk_field) == field_value)

        if await self._count(session, statement) > 1:
            raise ValueError('Multiple objects returned')
        result = await session.scalar(statement, **kwargs)

        if result is None:
            self.exceptions.not_found_exception(field_value)
        return result

    async def create(self, session: AsyncSession, data: dict[str, Any] = None) -> Table:
        """Create object."""
        self.validate_async_session(session)
        if data is None:
            data = {}
        entry = self.model(**data)
        try:
            session.add(entry)
            await session.commit()
        except IntegrityError as e:
            await session.rollback()
            self.exceptions.integrity_error(str(e.orig))

        await session.refresh(entry)
        return entry

    async def delete(self, session: AsyncSession, field_value: Any, **kwargs) -> Response:
        """Remove object."""
        if self.pk_field is None:
            self.exceptions.attribute_error('pk_field')
        self.validate_async_session(session)
        self.validate_field_value(field_value)

        statement = delete(self.model).where(getattr(self.model, self.pk_field) == field_value)
        await session.execute(statement, **kwargs)
        await session.commit()
        return Response(status_code=status.HTTP_204_NO_CONTENT)

    async def update_one(
            self,
            session: AsyncSession,
            field_value: Any,
            data: dict[str, Any] = None,
            synchronize_session='fetch',
            **kwargs,
    ) -> None:
        """Update object."""
        self.validate_async_session(session)
        self.validate_field_value(field_value)
        if self.pk_field is None:
            self.exceptions.attribute_error('pk_field')
        if data is None or not data:
            self.exceptions.empty_update_error()

        statement = update(self.model) \
            .where(getattr(self.model, self.pk_field) == field_value) \
            .values(**data) \
            .execution_options(synchronize_session=synchronize_session)
        try:
            await session.execute(statement, **kwargs)
        except IntegrityError as e:
            await session.rollback()
            self.exceptions.integrity_error(str(e.orig))
        else:
            await session.commit()



from functools import cached_property
from typing import Any, Type, TypeVar, Union

from sqlalchemy import select
from sqlalchemy.sql import Select

from fastapi_views.ext.sql_alchemy.async_api import AsyncAPI
from fastapi_views.ext.sql_alchemy.sync_api import SyncAPI

from ...types import Table
from ...views.mixins import (
    AbstractAPIMixin,
    BaseAPICreateMixin,
    BaseAPIDestroyMixin,
    BaseAPIDetailMixin,
    BaseAPIListMixin,
    BaseAPIUpdateMixin
)
from .core import BaseAPIView

API_CLS = TypeVar('API_CLS', bound=BaseAPIView)


class APIMixin(AbstractAPIMixin):
    """Base API mixin."""

    model: Type[Table] = None
    """A database table model class."""
    pk_field: Union[str, None] = None
    """Unique field like model id."""
    async_api: bool = False
    """When set to True then will be used async repository."""

    @cached_property
    def _repository_kwargs(self) -> dict:
        return dict(
            model=self.model,
            pk_field=self.pk_field,
            paginate_by=self.paginate_by,
            pagination_strategy=self.pagination_strategy,
        )

    @property
    def _sync_instance(self) -> SyncAPI:
        return SyncAPI[self.model](**self._repository_kwargs)

    @property
    def _async_instance(self) -> AsyncAPI:
        return AsyncAPI[self.model](**self._repository_kwargs)

    def _get_repository(self) -> API_CLS:
        """Get repository API cls."""
        return self._sync_instance if not self.async_api else self._async_instance

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if not isinstance(self.async_api, bool):
            raise TypeError(f'{self.async_api} expected to be boolean value.')

        # assigned after, so we can reference to api_view in both of the methods if needed
        self.api_view.pagination_kwargs = self.get_pagination_kwargs()
        self.api_view.statement = self.get_statement()

    def get_statement(self) -> Select:
        """Return sqlalchemy orm statement."""


class APIListMixin(APIMixin, BaseAPIListMixin):
    """Retrieve object list mixin."""

    pk_field = None

    def get_statement(self) -> Select:
        return select(self.model)

    def get_pagination_kwargs(self) -> dict[str, Any]:
        kw = super().get_pagination_kwargs()
        kw['ordering'] = ['id']
        return kw


class APIDetailMixin(APIMixin, BaseAPIDetailMixin):
    """Retrieve object mixin."""

    pk_field = 'id'


class APIUpdateMixin(APIMixin, BaseAPIUpdateMixin):
    """Update objects mixin."""

    pk_field = 'id'


class APICreateMixin(APIMixin, BaseAPICreateMixin):
    """Post objects mixin."""

    pk_field = None


class APIDestroyMixin(APIMixin, BaseAPIDestroyMixin):
    """Delete objects mixin."""

    pk_field = 'id'



import abc
from typing import Any, Mapping

from fastapi import HTTPException, status
from fastapi.requests import Request
from pydantic.dataclasses import dataclass
from pymongo.client_session import ClientSession
from pymongo.database import Database

from fastapi_views.ext.pymongo.pagination.core import paginate_api
from fastapi_views.settings import BaseConfig
from fastapi_views.types import Record
from fastapi_views.views.core import BaseRepositoryView, RepositoryExceptions


class APIErrorHandlers:
    exceptions: RepositoryExceptions
    statement: Mapping[str, Any]
    model: str

    def statement_type_error(self) -> None:
        raise TypeError('Statement must be type of dict')

    def session_type_error(self) -> None:
        raise TypeError('Session must be type of pymongo ClientSession')

    def invalid_document(self, error) -> None:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=error,
        )

    def validate_session(self, session: ClientSession) -> None:
        if session is not None and not isinstance(session, ClientSession):
            self.session_type_error()

    def validate_statement(self) -> None:
        if self.statement is None:
            self.exceptions.attribute_error('statement')
        if not isinstance(self.statement, dict):
            self.statement_type_error()

    def validate_field_value(self, unique: Any) -> None:
        if unique is None:
            self.exceptions.attribute_error('pk')

    def not_found_exception(self, unique: Any):
        """Object not found exception."""
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f'{self.model} with <{type(unique)}:{unique}> does not exists.'
        )


@dataclass(config=BaseConfig)
class BaseAPIView(BaseRepositoryView, APIErrorHandlers, abc.ABC):
    """
    Arguments:
        ** model ** : mongodb collection name
    """
    pk_field: Any = None
    statement: Mapping[str, Any] = None

    def get_all_with_pagination(
            self,
            db: Database,
            request: Request,
            session: ClientSession = None,
            **kwargs,
    ):
        """Returns multiple objects using pagination."""
        self.validate_session(session)
        self.validate_pagination(request)
        self.pagination_kwargs['collection'] = getattr(db, self.model)

        paginate_response = paginate_api(
            statement=self.statement,
            request=request,
            session=session,
            paginate_by=self.paginate_by,
            strategy=self.pagination_strategy,
            kwargs=self.pagination_kwargs,
        )
        return paginate_response

    @abc.abstractmethod
    def get_all(
            self,
            db: Database,
            session: ClientSession = None,
            limit: int = 100,
            skip: int = 0,
            **kwargs) -> list[Record]:
        """Returns multiple objects."""

    @abc.abstractmethod
    def get_detail(
            self,
            db: Database,
            session: ClientSession = None,
            field_value: Any = None,
            **kwargs,
    ):
        """Return single object."""

    @abc.abstractmethod
    def create(
            self,
            db: Database,
            document: Mapping[str, Any],
            session: ClientSession = None,
            **kwargs,
    ):
        """Perform object creation."""

    @abc.abstractmethod
    def delete(
            self,
            db: Database,
            field_value: Any,
            session: ClientSession = None,
            **kwargs,
    ):
        """Remove object."""

    @abc.abstractmethod
    def update_one(
            self,
            db: Database,
            update_doc: Mapping[str, Any],
            field_value: Any = None,
            session: ClientSession = None,
            document: Mapping[str, Any] = None,
            **kwargs,
    ):
        """ Perform single object update."""



import pytest  # noqa: F401

from .fixtures.requests import request_fixture  # noqa: F401
from .fixtures.sessions import motor_client_session  # noqa: F401
from .fixtures.sessions import (
    client_session,
    motor_collection,
    motor_database,
    pymongo_database
)



import abc
# from api.base import BaseAPI
from typing import Any, Union

from fastapi.requests import Request
from pydantic.dataclasses import dataclass
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session
from sqlalchemy.sql import Select

from fastapi_views.ext.sql_alchemy.pagination.core import paginate_api
from fastapi_views.settings import BaseConfig
from fastapi_views.types import Record
from fastapi_views.views.core import BaseRepositoryView, RepositoryExceptions


class APIErrorHandlers:
    exceptions: RepositoryExceptions
    statement: Select

    def statement_type_error(self) -> None:
        raise TypeError('Statement must be type of sqlalchemy.sql Select')

    def session_type_error(self) -> None:
        raise TypeError('Session must be type of sqlalchemy.orm Session')

    def async_session_type_error(self) -> None:
        raise TypeError('Session must be type of sqlalchemy.ext.asyncio AsyncSession')

    def validate_session(self, session: Session) -> None:
        if session is None:
            self.exceptions.attribute_error('session')
        if not isinstance(session, Session):
            self.session_type_error()

    def validate_async_session(self, session: AsyncSession) -> None:
        if session is None:
            self.exceptions.attribute_error('session')
        if not isinstance(session, AsyncSession):
            self.async_session_type_error()

    def validate_statement(self) -> None:
        if self.statement is None:
            self.exceptions.attribute_error('statement')
        if not isinstance(self.statement, Select):
            self.statement_type_error()

    def validate_field_value(self, unique: Any) -> None:
        if unique is None:
            self.exceptions.attribute_error('pk')


@dataclass(config=BaseConfig)
class BaseAPIView(BaseRepositoryView, APIErrorHandlers, abc.ABC):
    pk_field: Any = None
    statement: Select = None

    def get_all_with_pagination(self, request: Request, session: Session):
        """Returns multiple objects using pagination."""
        self.validate_session(session)
        self.validate_pagination(request)

        paginate_response = paginate_api(
            statement=self.statement,
            request=request,
            session=session,
            paginate_by=self.paginate_by,
            strategy=self.pagination_strategy,
            kwargs=self.pagination_kwargs,
        )
        return paginate_response

    @abc.abstractmethod
    def get_all(
            self,
            session: Union[Session, AsyncSession],
            limit: int = 100,
            offset: int = 0,
            **kwargs,
    ) -> list[Record]:
        """Returns multiple objects."""

    @abc.abstractmethod
    def get_detail(self, session: Union[Session, AsyncSession], field_value: Any, **kwargs):
        """Return single object."""

    @abc.abstractmethod
    def create(self, session: Union[Session, AsyncSession], data: dict[str, Any]):
        """Perform object creation."""

    @abc.abstractmethod
    def delete(self, session: Union[Session, AsyncSession], field_value: Any, **kwargs):
        """Remove object."""

    @abc.abstractmethod
    def update_one(
            self,
            session: Union[Session, AsyncSession],
            field_value: Any,
            data: dict[str, Any],
            synchronize_session='fetch',
            **kwargs,
    ):
        """
        Perform single object update.

        synchronize_session: check sqlalchemy doc for details.
        """



import datetime

import mongomock
from faker import Faker
from mongomock_motor import AsyncMongoMockClient
from sqlalchemy import text

from fastapi_views.tests._tmp_db import (
    Base,
    SimpleTestOrm,
    TestingSessionLocal,
    TestingSessionLocalAsync,
    async_engine,
    engine
)

faker = Faker()


class _PymongoBase:
    def __init__(self, count: int = 50):
        self.count = count
        self.collection = None

    def _create_random_documents(self):
        result = []
        for x in range(self.count):
            name = faker.first_name().lower()
            d = {
                'username': name,
                'email': f'{name}@example.com',
                'id': x + 1,
            }
            result.append(d)
        return result

    def create_documents(self):
        self.collection.insert_many(self._create_random_documents())


class PymongoDB(_PymongoBase):

    def __init__(self, count: int = 50):
        super().__init__(count)
        self.client = mongomock.MongoClient()
        self.db = self.client.get_database('test_db')

    def create_collection(self, name: str = 'test_collection'):
        self.db.create_collection(name)
        self.collection = getattr(self.db, name)


class MotorDB(_PymongoBase):

    def __init__(self, count: int = 50):
        super().__init__(count)
        self.client = AsyncMongoMockClient()
        self.db = self.client['test_db']
        self.collection = self.db['test_collection']

    async def create_documents(self):
        await self.collection.insert_many(self._create_random_documents())


class DatabaseSession:

    def __init__(self, number: int = 50):
        self._session = None
        self._engine = None
        self._number = number

    def init(self):
        self._engine = engine
        self._session = TestingSessionLocal()

    def create_all(self):
        Base.metadata.create_all(bind=engine)
        _now = datetime.datetime.now()
        for _id in range(0, self._number):
            t = _now + datetime.timedelta(minutes=_id)
            self._session.add(SimpleTestOrm(time_created=t))
        self._session.commit()

    def close_and_drop(self):
        self._session.close()
        Base.metadata.drop_all(bind=engine)

    @property
    def session(self):
        return self._session


class AsyncDatabaseSession:

    def __init__(self):
        self._session = None
        self._engine = None

    async def init(self):
        self._engine = async_engine
        self._session = TestingSessionLocalAsync()

    async def create_all(self):
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)
            ids = ', '.join(
                map(
                    lambda x: f'({x}, "{datetime.datetime.now()}")',
                    [x for x in range(1, 16)],
                ),
            )
            insert_txt = text(f"""
                INSERT INTO test_database (id, time_created)
                VALUES {ids}
            """)
            await conn.execute(insert_txt)

    async def delete_all(self):
        async with self._engine.begin() as conn:
            await conn.execute(text('DROP TABLE test_database'))

    @property
    def session(self):
        return self._session


class DatabaseSessionSameDateTime(AsyncDatabaseSession):

    async def create_all(self):
        _now = datetime.datetime.now()
        async with self._engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)
            ids = ', '.join(
                map(
                    lambda x: f'({x}, "{_now}")',
                    [x for x in range(0, 50)],
                ),
            )
            insert_txt = text(f"""
                INSERT INTO test_database (id, time_created)
                VALUES {ids}
            """)
            await conn.execute(insert_txt)



import abc
from typing import Any, Type, TypeVar, Union

from ..pagination.core import PaginationBase
from .core import BaseRepositoryView

API_CLS = TypeVar('API_CLS', bound=BaseRepositoryView)


class BaseAPIMixin:
    """Base API mixin."""

    paginate_by: Union[int, None] = None
    """Page objects limit."""
    pagination_strategy: Union[Type[PaginationBase], None] = None
    """As default set to limit offset."""
    model: Any = None

    def __init__(self, *args, **kwargs) -> None:
        attrs = kwargs.pop('attrs', {})
        self.args = args
        self.kwargs = kwargs
        self._api = None

        for key, value in attrs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise AttributeError(f'{key} is not a valid attribute')

    @property
    def api_view(self) -> API_CLS:
        return self._api

    @api_view.setter
    def api_view(self, value):
        if not isinstance(value, BaseRepositoryView):
            raise TypeError('Invalid type of api_view.')
        self._api = value

    def get_pagination_kwargs(self) -> dict[str, Any]:
        """
        Return pagination kwargs.
        Used only in cursor pagination.
        """


class AbstractAPIMixin(BaseAPIMixin, abc.ABC):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.api_view = self._get_repository()

    @abc.abstractmethod
    def _get_repository(self) -> API_CLS:
        """Get repository API cls."""


class BaseAPIListMixin(BaseAPIMixin):
    """Retrieve object list mixin."""

    def get_all(self, *args, **kwargs):
        return self.api_view.get_all(*args, **kwargs)

    def get_all_with_pagination(self, *args, **kwargs):
        return self.api_view.get_all_with_pagination(*args, **kwargs)

    def get_pagination_kwargs(self) -> dict[str, Any]:
        return {
            'model': self.model,
            'ordering': [],
            'cursor_prefixes': ['next__', 'prev__']
        }


class BaseAPIDetailMixin(BaseAPIMixin):
    """Retrieve object mixin."""

    def get_detail(self, *args, **kwargs):
        return self.api_view.get_detail(*args, **kwargs)


class BaseAPIUpdateMixin(BaseAPIMixin):
    """Update objects mixin."""

    def update_one(self, *args, **kwargs):
        return self.api_view.update_one(*args, **kwargs)


class BaseAPICreateMixin(BaseAPIMixin):
    """Post objects mixin."""

    def create(self, *args, **kwargs):
        return self.api_view.create(*args, **kwargs)


class BaseAPIDestroyMixin(BaseAPIMixin):
    """Delete objects mixin."""

    def delete(self, *args, **kwargs):
        return self.api_view.delete(*args, **kwargs)



from typing import Awaitable
from unittest.mock import patch

import pytest
import pytest_asyncio

from fastapi_views.ext.motor.pagination.core import (
    AsyncPaginationCursor,
    AsyncPaginationLimitOffset,
    AsyncPaginatorAPI,
    async_paginate_api
)
from fastapi_views.pagination.annotations import CursorResponse
from fastapi_views.utils import encode_value


@pytest_asyncio.fixture
async def async_limit_offset_api(request_fixture, request, motor_collection) -> dict:
    request_fixture.query_params = {
        'page': request.param,
        'limit': 3,
    }
    return await async_paginate_api(
        statement={},
        request=request_fixture,
        kwargs={
            'collection': motor_collection
        }
    )


@pytest_asyncio.fixture
@patch('fastapi_views.ext.motor.pagination.core.AsyncPaginationCursor._filter_query')
async def async_cursor_api(q, request_fixture, request, motor_collection) -> Awaitable[dict]:
    cursor, ordering, operator_expr, operator_value = request.param

    if cursor is not None:
        q.return_value = motor_collection.find(
            {'$expr': {f'${operator_expr}': ['$id', operator_value]}}
        )
    request_fixture.query_params = {
        'limit': 2,
        'cursor': cursor,
    }
    return await async_paginate_api(
        statement={},
        request=request_fixture,
        strategy=AsyncPaginationCursor,
        kwargs={
            'model': 'test_collection',
            'ordering': ordering,
            'cursor_prefixes': ['custom_next__', 'custom_prev__'],
            'collection': motor_collection
        }
    )


class TestPaginatorAPI:

    @patch('fastapi_views.ext.motor.pagination.core.AsyncPaginatorAPI.strategy_response')
    def test_init(self, response_mock, request_fixture):
        """Test if values are initialized as expected."""
        response_mock.return_value = {}

        paginator = AsyncPaginatorAPI(
            statement={},
            request=request_fixture,
            paginate_by=5
        )
        assert paginator.statement is not None
        assert paginator.request is not None
        assert paginator.session is None
        assert paginator.strategy is not None
        assert paginator.paginate_by is not None
        assert paginator.strategy == AsyncPaginationLimitOffset
        assert paginator.paginate_by == 5
        assert paginator.kwargs == {
            'model': None,
            'ordering': None,
            'cursor_prefixes': None,
        }


@pytest.mark.asyncio
class TestLimitOffset:

    @pytest.mark.parametrize(
        'async_limit_offset_api',
        [0, 1, 2, 3, 16],
        indirect=['async_limit_offset_api'],
    )
    async def test_strategy_response_common(
            self,
            async_limit_offset_api: dict,
    ):
        """Test pagination response the same common values in every page."""
        assert async_limit_offset_api['total'] == 50
        assert async_limit_offset_api['total_pages'] == 16
        assert async_limit_offset_api['page_limit'] == 3

    @pytest.mark.parametrize('async_limit_offset_api', [0], indirect=['async_limit_offset_api'])
    async def test_strategy_response(self, async_limit_offset_api: dict):
        """Test pagination response starting at page 0."""
        assert async_limit_offset_api['count'] == 3
        assert async_limit_offset_api['next_page'] == 'https://example.com?page=1&limit=3'
        assert async_limit_offset_api['previous_page'] is None
        assert async_limit_offset_api['last_page'] == 'https://example.com?page=16&limit=3'

    @pytest.mark.parametrize('async_limit_offset_api', [1], indirect=['async_limit_offset_api'])
    async def test_strategy_response_1(self, async_limit_offset_api: dict):
        """Test pagination response starting at page 1."""
        assert async_limit_offset_api['count'] == 3
        assert async_limit_offset_api['next_page'] == 'https://example.com?page=2&limit=3'
        assert async_limit_offset_api['previous_page'] == 'https://example.com?page=0&limit=3'
        assert async_limit_offset_api['last_page'] == 'https://example.com?page=16&limit=3'

    @pytest.mark.parametrize('async_limit_offset_api', [16], indirect=['async_limit_offset_api'])
    async def test_strategy_response_last(self, async_limit_offset_api: dict):
        """Test pagination response starting in last page."""
        assert async_limit_offset_api['count'] == 2
        assert async_limit_offset_api['next_page'] is None
        assert async_limit_offset_api['previous_page'] == 'https://example.com?page=15&limit=3'
        assert async_limit_offset_api['last_page'] == 'https://example.com?page=16&limit=3'


cursor_to_load_last_page = encode_value('custom_next__48&page__16').decode("utf-8")
cursor_to_load_last_page_reverse = encode_value('custom_next__3&page__4').decode("utf-8")


def assert_result(result: CursorResponse, next_cursor=None, prev_cursor=None):
    """
    Count is the same for all results because in our test we have 10 objects,
    which paginated by 2 gives equal number of objects per page.
    """
    assert result['count'] == 2
    assert result['page_limit'] == 2
    assert result['first_page'] == 'https://example.com?limit=2'
    if next_cursor is not None:
        next_page_url = f'https://example.com?limit=2&cursor={next_cursor.decode("utf-8")}'
        assert result['next_page'] == next_page_url
    else:
        assert result['next_page'] is None

    if prev_cursor is not None:
        prev_url = f'https://example.com?limit=2&cursor={prev_cursor.decode("utf-8")}'
        assert result['previous_page'] == prev_url
    else:
        assert result['previous_page'] is None


class TestCursorAscendingOrder:
    """We assume in tests that we have 50 objects ordered by id paginated by 2."""

    @staticmethod
    def _get_doc_id(doc: dict):
        return doc.get('id')

    @staticmethod
    def assert_common(result):
        assert result['count'] == 2
        assert result['page_limit'] == 2
        assert result['first_page'] == 'https://example.com?limit=2'

    @pytest.mark.parametrize(
        'async_cursor_api',
        [(None, ['id'], None, None)],
        indirect=['async_cursor_api'],
    )
    def test_strategy_response(self, async_cursor_api):
        """Test pagination response."""
        next_page_cursor = encode_value('custom_next__2&page__1')
        assert_result(async_cursor_api, next_page_cursor)
        api_ids = list(map(self._get_doc_id, async_cursor_api['results']))
        assert api_ids == [1, 2]

    @pytest.mark.parametrize(
        'async_cursor_api',
        [('Y3VzdG9tX25leHRfXzImcGFnZV9fMQ==', ['id'], 'gt', 2)],
        indirect=['async_cursor_api'],
    )
    def test_strategy_response_with_cursor(self, async_cursor_api: dict):
        """Test pagination response 'at page 1'."""
        next_page_cursor = encode_value('custom_next__4&page__2')
        previous_page_cursor = encode_value('custom_prev__3&page__0')
        assert_result(async_cursor_api, next_page_cursor, previous_page_cursor)
        api_ids = list(map(self._get_doc_id, async_cursor_api['results']))
        assert api_ids == [3, 4]

    @pytest.mark.parametrize(
        'async_cursor_api',
        [(cursor_to_load_last_page, ['id'], 'gt', 48)],
        indirect=['async_cursor_api'],
    )
    def test_strategy_response_with_cursor_last_page(self, async_cursor_api: dict):
        """Test pagination response 'in last page'."""
        previous_page_cursor = encode_value('custom_prev__49&page__15')
        assert_result(async_cursor_api, prev_cursor=previous_page_cursor)

        api_ids = list(map(self._get_doc_id, async_cursor_api['results']))
        assert api_ids == [49, 50]


class TestCursorDescendingOrder:
    """We assume in tests that we have 50 objects ordered by id desc paginated by 2."""

    @staticmethod
    def _get_doc_id(doc: dict):
        return doc.get('id')

    @pytest.mark.parametrize(
        'async_cursor_api',
        [(None, ['-id'], None, None)],
        indirect=['async_cursor_api'],
    )
    def test_strategy_response(self, async_cursor_api: dict):
        """Test pagination response."""
        next_page_cursor = encode_value('custom_next__49&page__1')
        assert_result(async_cursor_api, next_page_cursor)

        api_ids = list(map(self._get_doc_id, async_cursor_api['results']))
        assert api_ids == [50, 49]

    @pytest.mark.parametrize(
        'async_cursor_api',
        [('Y3VzdG9tX25leHRfXzkmcGFnZV9fMQ==', ['-id'], 'lt', 49)],
        indirect=['async_cursor_api'],
    )
    def test_strategy_response_with_cursor(self, async_cursor_api: dict):
        """Test pagination response 'at page 1'."""
        next_page_cursor = encode_value('custom_next__47&page__2')
        previous_page_cursor = encode_value('custom_prev__48&page__0')
        assert_result(async_cursor_api, next_page_cursor, previous_page_cursor)

        api_ids = list(map(self._get_doc_id, async_cursor_api['results']))
        assert api_ids == [48, 47]

    @pytest.mark.parametrize(
        'async_cursor_api',
        [(cursor_to_load_last_page_reverse, ['-id'], 'lt', 3)],
        indirect=['async_cursor_api'],
    )
    def test_strategy_response_with_cursor_last_page(self, async_cursor_api: dict):
        """Test pagination response 'in last page'."""
        previous_page_cursor = encode_value('custom_prev__2&page__3')
        assert_result(async_cursor_api, prev_cursor=previous_page_cursor)

        api_ids = list(map(self._get_doc_id, async_cursor_api['results']))
        assert api_ids == [2, 1]



"""Tests for sqlalchemy SyncAPI and mixins."""

import pytest
from fastapi import HTTPException
from pydantic import ValidationError
from sqlalchemy import select
from sqlalchemy.sql import Select

from fastapi_views.ext.sql_alchemy.async_api import AsyncAPI
from fastapi_views.ext.sql_alchemy.mixins import (
    APICreateMixin,
    APIDestroyMixin,
    APIDetailMixin,
    APIListMixin,
    APIMixin,
    APIUpdateMixin
)
from fastapi_views.ext.sql_alchemy.pagination.core import PaginationCursor
from fastapi_views.ext.sql_alchemy.sync_api import SyncAPI
from fastapi_views.tests._db_sessions import DatabaseSession
from fastapi_views.tests._tmp_db import SimpleTestOrm

NUMBER_OF_OBJECTS = 50  # number of objects created by sync db


@pytest.fixture
def session():
    db_session = DatabaseSession()
    db_session.init()
    db_session.create_all()
    yield db_session.session
    db_session.close_and_drop()


@pytest.fixture
def api(request):
    model, pk_field, _, paginate_by, strategy, kwargs, statement = request.param
    return SyncAPI(
        model=model,
        pk_field=pk_field,
        paginate_by=paginate_by,
        pagination_strategy=strategy,
        pagination_kwargs=kwargs,
        statement=statement,
    )


@pytest.fixture
def base_mixin_instance(request):
    _async_api = request.param
    args = ['a', 1, 'c', {'d'}]
    kwargs = {'a': 1}

    class BaseView(APIMixin):
        model = SimpleTestOrm
        async_api = _async_api

    return BaseView(*args, **kwargs)


@pytest.fixture
def list_mixin_instance(request):
    _model, _, _paginate_by, _strategy = request.param

    class ListView(APIListMixin):
        model = _model
        paginate_by = _paginate_by
        pagination_strategy = _strategy

    return ListView()


@pytest.fixture
def list_mixin_instance_override_get_statement(request):
    _model = request.param

    class ListView(APIListMixin):
        model = _model

        def get_statement(self) -> Select:
            return select(self.model).where(self.model.id > 48)

    return ListView()


@pytest.fixture
def detail_mixin_instance(request):
    _model = request.param

    class DetailView(APIDetailMixin):
        model = _model

    return DetailView()


@pytest.fixture
def detail_mixin_instance_override_get_statement(request):
    _model = request.param

    class DetailView(APIDetailMixin):
        model = _model

        def get_statement(self) -> Select:
            return select(self.model).where(self.model.id == 10)

    return DetailView()


@pytest.fixture
def create_mixin_instance(request):
    _model = request.param

    class CreateView(APICreateMixin):
        model = _model

    return CreateView()


@pytest.fixture
def update_mixin_instance(request):
    _model = request.param

    class UpdateView(APIUpdateMixin):
        model = _model

    return UpdateView()


@pytest.fixture
def delete_mixin_instance(request):
    _model = request.param

    class DeleteView(APIDestroyMixin):
        model = _model

    return DeleteView()


class TestSyncAPI:
    """Test SyncAPI cls."""

    @pytest.mark.parametrize('test_input', [None, ''])
    def test_init_model_raise_value_error(self, test_input):
        """Test if value error occurs on given inputs."""
        with pytest.raises(ValueError):
            assert SyncAPI(model=test_input)

    @pytest.mark.parametrize('test_input', ['not_cls', 1, True, {'a': 'b'}, {1}])
    def test_init_model_raise_validation_error(self, test_input):
        """Test if pydantic validation error occurs on given inputs."""
        with pytest.raises(ValidationError):
            assert SyncAPI(model=test_input)

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, None, None, None, None)],
        indirect=['api'],
    )
    def test_get_all_missing_session(self, api):
        """Test session attr in get_all method."""
        with pytest.raises(AttributeError):
            api.get_all(None)
        with pytest.raises(TypeError):
            api.get_all(False)
        with pytest.raises(TypeError):
            api.get_all('not_session')

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, None, None, None, None)],
        indirect=['api'],
    )
    def test_get_all_statement_error(self, session, api):
        """Test statement attr in get_all method."""
        with pytest.raises(AttributeError):
            api.get_all(session=session)
        with pytest.raises(TypeError):
            api.statement = 'invalid'
            api.get_all(session=session)

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, None, None, None, select(SimpleTestOrm))],
        indirect=['api'],
    )
    def test_get_all(self, session, api):
        """Test if returns all created objects."""
        assert len(api.get_all(session=session)) == NUMBER_OF_OBJECTS

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, None, None, None, None)],
        indirect=['api'],
    )
    def test_count(self, session, api):
        """Test if returns correct number related to statement."""
        assert api._count(
            session=session,
            statement=select(SimpleTestOrm).where(SimpleTestOrm.id < 25),
        ) == 24

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, None, None, None, None)],
        indirect=['api'],
    )
    def test_count_missing_attrs(self, session, api):
        """Test attrs in _count method."""
        with pytest.raises(TypeError):
            api._count()

        with pytest.raises(TypeError):
            api._count(session=session)

        with pytest.raises(TypeError):
            api._count(statement=select(SimpleTestOrm))

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, 'id', None, None, None, None, None)],
        indirect=['api'],
    )
    def test_get_detail_statement_none_fk_fields_none(self, session, api):
        """Test get_detail method when statement and fk fields are not provided."""
        result = api.get_detail(
            session=session,
            field_value=1,
        )
        assert result.id == 1
        assert type(result).__name__ == 'SimpleTestOrm'

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, None, None, None, select(SimpleTestOrm).where(
            SimpleTestOrm.id == 5,
        ))],
        indirect=['api'],
    )
    def test_get_detail_statement_not_none_fk_fields_none(self, session, api):
        """Test get_detail method when fk fields are not provided."""
        result = api.get_detail(
            session=session,
        )
        assert result.id == 5
        assert type(result).__name__ == 'SimpleTestOrm'

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, None, None, None, select(SimpleTestOrm))],
        indirect=['api'],
    )
    def test_get_detail_multiple_objects_returned(self, session, api):
        """Test get_detail method when multiple objects are returned."""
        with pytest.raises(ValueError):
            api.get_detail(
                session=session,
            )

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, 'id', None, None, None, None, None)],
        indirect=['api'],
    )
    def test_get_detail_not_found(self, session, api):
        """Test get_detail method when object is not found."""
        with pytest.raises(HTTPException):
            api.get_detail(
                session=session,
                field_value=51,
            )

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, None, None, None, None)],
        indirect=['api'],
    )
    def test_get_detail_missing_attrs(self, session, api):
        """Test get_detail attrs."""
        with pytest.raises(TypeError):
            api.get_detail()

        with pytest.raises(AttributeError):
            api.get_detail(
                session=None,
                field_value=10,
            )
        with pytest.raises(AttributeError):
            api.get_detail(
                session=session,
                field_value=None,
            )

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, None, None, None, None)],
        indirect=['api'],
    )
    def test_create(self, session, api):
        """Test create method when new object is created."""
        result = api.create(
            session=session
        )
        assert result.id > NUMBER_OF_OBJECTS
        assert type(result).__name__ == 'SimpleTestOrm'

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, None, None, None, None)],
        indirect=['api'],
    )
    def test_create_attrs(self, session, api):
        """Test create method attrs."""
        with pytest.raises(AttributeError):
            api.create(
                session=None,
            )
        with pytest.raises(TypeError):
            api.create(
                session='not_session',
            )
        with pytest.raises(TypeError):
            api.create(
                session=2,
            )
        with pytest.raises(TypeError):
            api.create(
                session=True,
            )
        with pytest.raises(TypeError):
            api.create(
                session=session,
                data={
                    'invalid': 'invalid'
                }
            )

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, None, None, None, None)],
        indirect=['api'],
    )
    def test_create_integrity_error(self, session, api):
        """Test create method integrity error."""
        with pytest.raises(HTTPException):
            api.create(
                session=session,
                data={
                    'id': 'invalid'
                }
            )

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, 'id', None, None, None, None, None)],
        indirect=['api'],
    )
    def test_delete(self, session, api):
        """Test delete method."""
        assert api._count(session, select(SimpleTestOrm)) == NUMBER_OF_OBJECTS
        api.delete(
            session=session,
            field_value=50,
        )
        # check if delete removed object
        with pytest.raises(HTTPException):
            api.get_detail(session, 50)

        assert api._count(session, select(SimpleTestOrm)) == NUMBER_OF_OBJECTS - 1

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, 'id', None, None, None, None, None)],
        indirect=['api'],
    )
    def test_delete_missing_attrs(self, session, api):
        """Test delete method attrs."""
        with pytest.raises(TypeError):
            api.delete()
        with pytest.raises(AttributeError):
            api.delete(session=None, field_value=2)

        with pytest.raises(TypeError):
            api.delete(session='not_valid_type', field_value=2)

        api.pk_field = None
        with pytest.raises(AttributeError):
            api.delete(session=session, field_value=2)

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, 'id', None, None, None, None, None)],
        indirect=['api'],
    )
    def test_update_one(self, session, api):
        """Test update_one method."""
        api.update_one(
            session=session,
            field_value=1,
            data={
                'id': 100
            }
        )
        assert api.get_detail(session, 100) is not None
        with pytest.raises(HTTPException):
            api.get_detail(session, 1)

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, 'id', None, None, None, None, None)],
        indirect=['api'],
    )
    def test_update_one_missing_attrs(self, session, api):
        """Test update_one method attrs."""
        with pytest.raises(TypeError):
            api.update_one()
        with pytest.raises(AttributeError):
            api.update_one(session=None, field_value=2)

        with pytest.raises(TypeError):
            api.update_one(session='not_valid_type', field_value=2)

        with pytest.raises(HTTPException):
            api.update_one(session=session, field_value=2)

        api.pk_field = None
        with pytest.raises(AttributeError):
            api.update_one(session=session, field_value=2, data={'id': 100})

        api.pk_field = 'id'
        with pytest.raises(HTTPException):
            api.update_one(session=session, field_value=2, data={'id': 'invalid'})

        assert api.get_detail(session, 2).id == 2

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, 4, None, None, select(SimpleTestOrm))],
        indirect=['api'],
    )
    def test_get_all_with_pagination_limit_offset(self, session, api, request_fixture):
        """Test get_all limit, offset pagination."""
        request_fixture.query_params = {}
        response = api.get_all_with_pagination(
            request=request_fixture,
            session=session,
        )
        assert response is not None
        assert isinstance(response, dict)
        assert response['count'] == 4
        assert response['page_limit'] == 4
        assert response['total'] == NUMBER_OF_OBJECTS

        # last page
        request_fixture.query_params = {
            'page': 12,
        }
        response = api.get_all_with_pagination(
            request=request_fixture,
            session=session,
        )
        assert response is not None
        assert isinstance(response, dict)
        assert response['count'] == 2
        assert response['next_page'] is None

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, 4, PaginationCursor, {
            'model': SimpleTestOrm,
            'ordering': ['id'],
            'cursor_prefixes': ['next__', 'prev__']
        }, select(SimpleTestOrm).where(SimpleTestOrm.id > 48))],
        indirect=['api'],
    )
    def test_get_all_with_pagination_cursor(self, session, api, request_fixture):
        """Test get_all cursor pagination."""
        request_fixture.query_params = {}
        response = api.get_all_with_pagination(
            request=request_fixture,
            session=session,
        )
        assert response is not None
        assert isinstance(response, dict)
        assert response['count'] == 2
        assert response['page_limit'] == 4
        assert response['next_page'] is None
        assert response['previous_page'] is None

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, 4, PaginationCursor, {
            'model': SimpleTestOrm,
            'ordering': ['time_created'],
            'cursor_prefixes': ['next__', 'prev__']
        }, select(SimpleTestOrm).where(SimpleTestOrm.id > 43))],
        indirect=['api'],
    )
    def test_get_all_with_pagination_cursor_datetime_ordering(self, session, api, request_fixture):
        """Test get_all cursor pagination using datetime field in ordering."""
        request_fixture.query_params = {}
        assert api.pagination_kwargs['ordering'] == ['time_created']
        response = api.get_all_with_pagination(
            request=request_fixture,
            session=session,
        )
        assert response is not None
        assert isinstance(response, dict)
        assert response['next_page'] is not None
        assert response['previous_page'] is None
        assert response['results'][0].id == 44  # first found
        assert response['count'] == 4

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, 5, PaginationCursor, {
            'model': SimpleTestOrm,
            'ordering': ['-time_created'],
            'cursor_prefixes': ['next__', 'prev__']
        }, select(SimpleTestOrm).where(SimpleTestOrm.id > 43))],
        indirect=['api'],
    )
    def test_get_all_with_pagination_cursor_datetime_ordering_reverted(
            self,
            session,
            api,
            request_fixture,
    ):
        """
        Test get_all cursor pagination using descending datetime field in ordering.

        In case the time hour, minute and second would be the same for all objects from statement,
        then id found would be equal to 44, because it would be no difference in ordering.

        This case is showed in test_async_api.py file along with ordering with 2 fields.
        """
        request_fixture.query_params = {}
        assert api.pagination_kwargs['ordering'] == ['-time_created']
        response = api.get_all_with_pagination(
            request=request_fixture,
            session=session,
        )
        assert response['results'][0].id == 50  # first found

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, 4, None, None, select(SimpleTestOrm))],
        indirect=['api'],
    )
    def test_get_all_with_pagination_missing_attrs(self, session, api, request_fixture):
        """Test get_all pagination attrs."""
        with pytest.raises(TypeError):
            api.get_all_with_pagination()
        with pytest.raises(AttributeError):
            api.get_all_with_pagination(session=None, request=request_fixture)

        with pytest.raises(AttributeError):
            api.get_all_with_pagination(session=session, request=None)
        with pytest.raises(TypeError):
            api.get_all_with_pagination(session=session, request='2')

        with pytest.raises(TypeError):
            api.get_all_with_pagination(session='not_valid_type', request=request_fixture)

    @pytest.mark.parametrize(
        'api',
        [(SimpleTestOrm, None, None, None, None, None, select(SimpleTestOrm))],
        indirect=['api'],
    )
    def test_get_all_with_pagination_invalid_paginate_by(self, session, api, request_fixture):
        """Test get_all pagination, paginate_by attr."""
        with pytest.raises(AttributeError):
            api.get_all_with_pagination(session=session, request=request_fixture)

        api.paginate_by = False
        with pytest.raises(TypeError):
            api.get_all_with_pagination(session=session, request=request_fixture)

        api.paginate_by = 'str'
        with pytest.raises(TypeError):
            api.get_all_with_pagination(session=session, request=request_fixture)

        api.paginate_by = 0
        with pytest.raises(ValueError):
            api.get_all_with_pagination(session=session, request=request_fixture)

        api.paginate_by = -1
        with pytest.raises(ValueError):
            api.get_all_with_pagination(session=session, request=request_fixture)


class TestMixins:
    """Test mixin classes."""

    @pytest.mark.parametrize(
        'base_mixin_instance',
        [False, True],
        indirect=['base_mixin_instance'],
    )
    def test_init(self, base_mixin_instance: APIMixin):
        """Test base mixin initialized as expected."""
        assert base_mixin_instance.args == ('a', 1, 'c', {'d'})
        assert base_mixin_instance.kwargs == {'a': 1}
        assert base_mixin_instance.api_view.statement is None
        assert base_mixin_instance.api_view.pagination_kwargs is None

    @pytest.mark.parametrize('input_value', ['some_str', 0, 1, {'a': 'b'}, {1}])
    def test_init_async_api_type_error(self, input_value):
        """Test base mixin async_api attr."""
        with pytest.raises(TypeError):
            class BaseView(APIMixin):
                model = SimpleTestOrm
                async_api = input_value

            BaseView()

    @pytest.mark.parametrize('base_mixin_instance', [False], indirect=['base_mixin_instance'])
    def test_init_async_false(self, base_mixin_instance: APIMixin):
        """Test if api_view is instance of sync API when async_api set to False."""
        assert isinstance(base_mixin_instance.api_view, SyncAPI)

    @pytest.mark.parametrize('base_mixin_instance', [True], indirect=['base_mixin_instance'])
    def test_init_async_true(self, base_mixin_instance: APIMixin):
        """Test if api_view is instance of async API when async_api set to True."""
        assert isinstance(base_mixin_instance.api_view, AsyncAPI)

    @pytest.mark.parametrize(
        'list_mixin_instance',
        [(SimpleTestOrm, None, None, None)],
        indirect=['list_mixin_instance'],
    )
    def test_list_mixin(self, session, list_mixin_instance):
        """Test list mixin."""
        assert len(list_mixin_instance.get_all(session=session)) == NUMBER_OF_OBJECTS

        assert isinstance(list_mixin_instance.api_view.statement, Select)
        assert list_mixin_instance.async_api is False
        assert list_mixin_instance.api_view.pagination_kwargs == {
            'model': SimpleTestOrm,
            'ordering': ['id'],
            'cursor_prefixes': ['next__', 'prev__']
        }

        with pytest.raises(AttributeError):
            list_mixin_instance.get_detail()
        with pytest.raises(AttributeError):
            list_mixin_instance.create()
        with pytest.raises(AttributeError):
            list_mixin_instance.delete()
        with pytest.raises(AttributeError):
            list_mixin_instance.update_one()

    @pytest.mark.parametrize(
        'list_mixin_instance',
        [(SimpleTestOrm, None, 10, None)],
        indirect=['list_mixin_instance'],
    )
    def test_list_mixin_paginate_by(self, session, list_mixin_instance, request_fixture):
        """Test list mixin with pagination."""
        request_fixture.query_params = {}

        assert list_mixin_instance.paginate_by == 10

        response = list_mixin_instance.get_all_with_pagination(
            session=session,
            request=request_fixture,
        )
        assert response['count'] == 10
        assert response['total_pages'] == NUMBER_OF_OBJECTS / 10 - 1

    @pytest.mark.parametrize(
        'list_mixin_instance',
        [(SimpleTestOrm, None, 10, PaginationCursor)],
        indirect=['list_mixin_instance'],
    )
    def test_list_mixin_cursor_strategy(self, session, list_mixin_instance, request_fixture):
        """Test list mixin pagination with cursor strategy."""
        request_fixture.query_params = {}

        response = list_mixin_instance.get_all_with_pagination(
            session=session,
            request=request_fixture,
        )
        assert isinstance(response, dict)
        assert list_mixin_instance.pagination_strategy == PaginationCursor

    @pytest.mark.parametrize(
        'detail_mixin_instance',
        [SimpleTestOrm],
        indirect=['detail_mixin_instance'],
    )
    def test_detail_mixin(self, session, detail_mixin_instance):
        """Test detail mixin."""
        response = detail_mixin_instance.get_detail(session=session, field_value=1)
        assert response.id == 1
        assert type(response).__name__ == 'SimpleTestOrm'

        assert detail_mixin_instance.api_view.statement is None
        assert detail_mixin_instance.async_api is False
        assert detail_mixin_instance.api_view.pagination_kwargs is None

        with pytest.raises(AttributeError):
            detail_mixin_instance.get_all()
        with pytest.raises(AttributeError):
            detail_mixin_instance.get_all_with_pagination()
        with pytest.raises(AttributeError):
            detail_mixin_instance.create()
        with pytest.raises(AttributeError):
            detail_mixin_instance.delete()
        with pytest.raises(AttributeError):
            detail_mixin_instance.update_one()

    @pytest.mark.parametrize(
        'list_mixin_instance_override_get_statement',
        [SimpleTestOrm],
        indirect=['list_mixin_instance_override_get_statement'],
    )
    def test_list_mixin_get_statement(self, session, list_mixin_instance_override_get_statement):
        """Test list mixin override get_statement method."""
        response = list_mixin_instance_override_get_statement.get_all(session=session)
        assert len(response) == 2

    @pytest.mark.parametrize(
        'detail_mixin_instance_override_get_statement',
        [SimpleTestOrm],
        indirect=['detail_mixin_instance_override_get_statement'],
    )
    def test_detail_mixin_get_statement(
            self, session,
            detail_mixin_instance_override_get_statement,
    ):
        """Test detail mixin override get_statement method."""
        response = detail_mixin_instance_override_get_statement.get_detail(session=session)
        assert response.id == 10

    @pytest.mark.parametrize(
        'create_mixin_instance',
        [SimpleTestOrm],
        indirect=['create_mixin_instance'],
    )
    def test_create_mixin(self, session, create_mixin_instance):
        """Test create mixin."""
        response = create_mixin_instance.create(session=session)
        assert response.id > NUMBER_OF_OBJECTS
        assert type(response).__name__ == 'SimpleTestOrm'

        assert create_mixin_instance.api_view.pk_field is None
        assert create_mixin_instance.api_view.statement is None
        assert create_mixin_instance.async_api is False
        assert create_mixin_instance.api_view.pagination_kwargs is None

        with pytest.raises(AttributeError):
            create_mixin_instance.get_all()
        with pytest.raises(AttributeError):
            create_mixin_instance.get_all_with_pagination()
        with pytest.raises(AttributeError):
            create_mixin_instance.get_detail()
        with pytest.raises(AttributeError):
            create_mixin_instance.delete()
        with pytest.raises(AttributeError):
            create_mixin_instance.update_one()

    @pytest.mark.parametrize(
        'update_mixin_instance',
        [SimpleTestOrm],
        indirect=['update_mixin_instance'],
    )
    def test_update_mixin(self, session, update_mixin_instance):
        """Test update mixin."""
        update_mixin_instance.update_one(session=session, field_value=1, data={'id': 100})

        api = SyncAPI(
            model=SimpleTestOrm,
            pk_field='id'
        )
        with pytest.raises(HTTPException):
            api.get_detail(session=session, field_value=1)
        api.get_detail(session=session, field_value=100)

        assert update_mixin_instance.api_view.pk_field is not None
        assert update_mixin_instance.api_view.statement is None
        assert update_mixin_instance.async_api is False
        assert update_mixin_instance.api_view.pagination_kwargs is None

        with pytest.raises(AttributeError):
            update_mixin_instance.get_all()
        with pytest.raises(AttributeError):
            update_mixin_instance.get_all_with_pagination()
        with pytest.raises(AttributeError):
            update_mixin_instance.get_detail()
        with pytest.raises(AttributeError):
            update_mixin_instance.create()
        with pytest.raises(AttributeError):
            update_mixin_instance.delete()

    @pytest.mark.parametrize(
        'delete_mixin_instance',
        [SimpleTestOrm],
        indirect=['delete_mixin_instance'],
    )
    def test_delete_mixin(self, session, delete_mixin_instance):
        """Test delete mixin."""
        delete_mixin_instance.delete(session=session, field_value=1)

        api = SyncAPI(
            model=SimpleTestOrm,
            pk_field='id'
        )
        with pytest.raises(HTTPException):
            api.get_detail(session=session, field_value=1)

        assert delete_mixin_instance.api_view.pk_field is not None
        assert delete_mixin_instance.api_view.statement is None
        assert delete_mixin_instance.async_api is False
        assert delete_mixin_instance.api_view.pagination_kwargs is None

        with pytest.raises(AttributeError):
            delete_mixin_instance.get_all()
        with pytest.raises(AttributeError):
            delete_mixin_instance.get_all_with_pagination()
        with pytest.raises(AttributeError):
            delete_mixin_instance.get_detail()
        with pytest.raises(AttributeError):
            delete_mixin_instance.create()
        with pytest.raises(AttributeError):
            delete_mixin_instance.update_one()



from typing import Any, Mapping, TypeVar, Union

from fastapi_views.ext.pymongo.sync_api import SyncAPI

from ...views.core import BaseRepositoryView
from ...views.mixins import (
    AbstractAPIMixin,
    BaseAPICreateMixin,
    BaseAPIDestroyMixin,
    BaseAPIDetailMixin,
    BaseAPIListMixin,
    BaseAPIUpdateMixin
)

API_CLS = TypeVar('API_CLS', bound=BaseRepositoryView)


class APIMixin(AbstractAPIMixin):
    """
    Base API mixin.

        ** Required Arguments **
            - model: name of the collection

        ** Optional Arguments **
            - pk_field: unique field like collection id
            - paginate_by: number of returned objects per page
            - pagination_strategy: PaginationLimitOffset ( default ) / PaginationCursor

    You can override get_statement and get_pagination_kwargs methods

    Examples:
            def get_pagination_kwargs(self) -> dict[str, Any]:
                kw = super().get_pagination_kwargs()
                kw['ordering'] = ['id']
                return kw

            def get_statement(self) -> Mapping[str, Any]:
                return {'name': 'john', 'age': {'$gt': 25}}
    """

    model: str = None
    """Name of the collection."""
    pk_field: Union[str, None] = None
    """Unique field like model id."""

    @property
    def _repository_kwargs(self) -> dict:
        return dict(
            model=self.model,
            pk_field=self.pk_field,
            paginate_by=self.paginate_by,
            pagination_strategy=self.pagination_strategy,
        )

    @property
    def _sync_instance(self) -> SyncAPI:
        return SyncAPI[self.model](**self._repository_kwargs)

    def _get_repository(self) -> API_CLS:
        """Get repository API cls."""
        return self._sync_instance

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # assigned after, so we can reference to api_view in both of the methods if needed
        self.api_view.pagination_kwargs = self.get_pagination_kwargs()
        self.api_view.statement = self.get_statement()

    def get_statement(self) -> Mapping[str, Any]:
        """Return mongodb raw statement."""


class APIListMixin(APIMixin, BaseAPIListMixin):
    """Retrieve object list mixin."""

    pk_field = None

    def get_statement(self) -> Mapping[str, Any]:
        return {}


class APIDetailMixin(APIMixin, BaseAPIDetailMixin):
    """Retrieve object mixin."""

    pk_field = '_id'


class APIUpdateMixin(APIMixin, BaseAPIUpdateMixin):
    """Update objects mixin."""

    pk_field = '_id'


class APICreateMixin(APIMixin, BaseAPICreateMixin):
    """Post objects mixin."""

    pk_field = None


class APIDestroyMixin(APIMixin, BaseAPIDestroyMixin):
    """Delete objects mixin."""

    pk_field = '_id'



from pathlib import Path

from setuptools import find_packages, setup

VERSION = '0.1.1'
DESCRIPTION = 'Fastapi view mixins'
# read the contents of your README file
this_directory = Path(__file__).parent
LONG_DESCRIPTION = (this_directory / "readme.md").read_text()

# Setting up
setup(
    name="fastapi-view-mixins",
    version=VERSION,
    author="Patryk Dąbrowski",
    author_email="tibiasportex@gmail.com",
    license_files=('LICENSE.txt',),
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    include_package_data=True,
    setup_requires=['setuptools_git >= 0.3'],
    install_requires=['fastapi', 'sqlalchemy', 'pymongo', 'motor'],
    exclude_package_data={'': ['.gitignore', 'requirements.txt']},

    keywords=['python', 'paginator', 'pagination', 'views'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



from typing import TypeVar

from fastapi_views.ext.pymongo.mixins import \
    APICreateMixin as PymongoAPICreateMixin
from fastapi_views.ext.pymongo.mixins import \
    APIDestroyMixin as PymongoAPIDestroyMixin
from fastapi_views.ext.pymongo.mixins import \
    APIDetailMixin as PymongoAPIDetailMixin
from fastapi_views.ext.pymongo.mixins import \
    APIListMixin as PymongoAPIListMixin
from fastapi_views.ext.pymongo.mixins import APIMixin as PymongoAPIMixin
from fastapi_views.ext.pymongo.mixins import \
    APIUpdateMixin as PymongoAPIUpdateMixin

from ...views.core import BaseRepositoryView
from .async_api import AsyncAPI

API_CLS = TypeVar('API_CLS', bound=BaseRepositoryView)


class APIMixin(PymongoAPIMixin):
    """
    Base API mixin.

        ** Required Arguments **
            - model: name of the collection

        ** Optional Arguments **
            - pk_field: unique field like collection id
            - paginate_by: number of returned objects per page
            - pagination_strategy: AsyncPaginationLimitOffset ( default ) / AsyncPaginationCursor

    You can override get_statement and get_pagination_kwargs methods

    Examples:
            def get_pagination_kwargs(self) -> dict[str, Any]:
                kw = super().get_pagination_kwargs()
                kw['ordering'] = ['id']
                return kw

            def get_statement(self) -> Mapping[str, Any]:
                return {'name': 'john', 'age': {'$gt': 25}}
    """

    @property
    def _async_instance(self) -> AsyncAPI:
        return AsyncAPI[self.model](**self._repository_kwargs)

    def _get_repository(self) -> API_CLS:
        """Get repository API cls."""
        return self._async_instance


class APIListMixin(APIMixin, PymongoAPIListMixin):
    """Retrieve object list mixin."""


class APIDetailMixin(APIMixin, PymongoAPIDetailMixin):
    """Retrieve object mixin."""


class APIUpdateMixin(APIMixin, PymongoAPIUpdateMixin):
    """Update objects mixin."""


class APICreateMixin(APIMixin, PymongoAPICreateMixin):
    """Post objects mixin."""


class APIDestroyMixin(APIMixin, PymongoAPIDestroyMixin):
    """Delete objects mixin."""



from sqlalchemy import Column, DateTime, Integer, create_engine
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import func

engine = create_engine("sqlite:///./test.db")
async_engine = create_async_engine("sqlite+aiosqlite:///./test.db")

TestingSessionLocal = sessionmaker(bind=engine)
TestingSessionLocalAsync = sessionmaker(bind=async_engine, class_=AsyncSession)

Base = declarative_base()


class SimpleTestOrm(Base):
    __tablename__ = 'test_database'

    id: int = Column(Integer, primary_key=True, index=True)
    time_created = Column(DateTime(timezone=True), server_default=func.now())



from typing import Any, Generic, TypeVar

from fastapi import status
from fastapi.responses import Response
from pydantic.dataclasses import dataclass
from sqlalchemy import delete, func, select, update
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session
from sqlalchemy.sql import Select

from fastapi_views.ext.sql_alchemy.core import BaseAPIView
from fastapi_views.settings import BaseConfig
from fastapi_views.types import Record

Table = TypeVar("Table")


@dataclass(config=BaseConfig)
class SyncAPI(BaseAPIView, Generic[Table]):
    """Sync API with default methods to get, create, update, delete objects."""

    @staticmethod
    def _count(session: Session, statement: Select) -> int:
        """Get number of documents for given statement."""
        count_statement = select(func.count()).select_from(statement.subquery())
        return session.scalar(count_statement)

    def get_all(
            self,
            session: Session,
            limit: int = 100,
            offset: int = 0,
            **kwargs,
    ) -> list[Record]:
        self.validate_session(session)
        self.validate_statement()
        return session.execute(self.statement.limit(limit).offset(limit * offset), **kwargs).all()

    def get_detail(self, session: Session, field_value: Any = None, **kwargs) -> Table:
        """Fetch object."""
        self.validate_session(session)

        if self.statement is not None:
            statement = self.statement
        else:
            self.validate_field_value(field_value)
            statement = select(self.model).where(getattr(self.model, self.pk_field) == field_value)

        if self._count(session, statement) > 1:
            raise ValueError('Multiple objects returned')
        result = session.scalar(statement, **kwargs)

        if result is None:
            self.exceptions.not_found_exception(field_value)
        return result

    def create(self, session: Session, data: dict[str, Any] = None) -> Table:
        """Create object."""
        self.validate_session(session)
        if data is None:
            data = {}
        entry = self.model(**data)
        try:
            session.add(entry)
            session.commit()
        except IntegrityError as e:
            session.rollback()
            self.exceptions.integrity_error(str(e.orig))

        session.refresh(entry)
        return entry

    def delete(self, session: Session, field_value: Any, **kwargs) -> Response:
        """Remove object."""
        if self.pk_field is None:
            self.exceptions.attribute_error('pk_field')
        self.validate_session(session)
        self.validate_field_value(field_value)

        statement = delete(self.model).where(getattr(self.model, self.pk_field) == field_value)
        session.execute(statement, **kwargs)
        session.commit()
        return Response(status_code=status.HTTP_204_NO_CONTENT)

    def update_one(
            self,
            session: Session,
            field_value: Any,
            data: dict[str, Any] = None,
            synchronize_session='fetch',
            **kwargs,
    ) -> None:
        """Update object."""
        self.validate_session(session)
        self.validate_field_value(field_value)
        if self.pk_field is None:
            self.exceptions.attribute_error('pk_field')
        if data is None or not data:
            self.exceptions.empty_update_error()

        statement = update(self.model) \
            .where(getattr(self.model, self.pk_field) == field_value) \
            .values(**data) \
            .execution_options(synchronize_session=synchronize_session)
        try:
            session.execute(statement, **kwargs)
        except IntegrityError as e:
            session.rollback()
            self.exceptions.integrity_error(str(e.orig))
        else:
            session.commit()



from unittest.mock import MagicMock

import pytest
from fastapi import HTTPException
from sqlalchemy.engine import Row

from .. import utils


def test_scalars():
    empty_response = {}
    assert utils.scalars(empty_response) == {}
    mocks = [MagicMock(spec=Row, return_value=(x, )) for x in range(0, 2)]
    response = {'results': [mocks[0], mocks[1]]}
    assert utils.scalars(response) == {
        'results': [mocks[0][0], mocks[1][0]],
    }
    assert utils.scalars(response, data_key='invalid') == response

    response = [(mocks[0],), (mocks[1],)]
    assert utils.scalars(response) == [mocks[0], mocks[1]]

    assert utils.scalars(response, data_key='invalid') == mocks


@pytest.mark.parametrize('test_input, expected', [('(?<=-).*', '=2a?cb'), ('^[^?]*', 'ran89-=2a')])
def test_regex_group(test_input, expected):
    search_value = 'ran89-=2a?cb'
    assert utils.get_regex_group(test_input, search_value) == expected

    with pytest.raises(ValueError):
        utils.get_regex_group('(?<=-)', '')


def test_encode_decode():
    value = 'test'
    assert utils.encode_value(value) == b'dGVzdA=='
    assert utils.decode_value('dGVzdA==') == value

    with pytest.raises(HTTPException):
        utils.decode_value('invalid')


def test_dict_to_query_string():
    assert utils.dict_to_query_string({'some_value': 1, 'next__': 'x'}) == 'some_value=1&next__=x'
